# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class ProtectionRule:
    NO_ACTION = 0
    MONITOR = 1
    BLOCK = 2
    BLOCK_AT_PERIMETER = 3
    PERMIT = 4
