# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dtm.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='dtm.proto',
  package='contrast.api.dtm',
  syntax='proto3',
  serialized_options=b'Z:bitbucket.org/contrastsecurity/go-speedracer-go/from_agent',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\tdtm.proto\x12\x10\x63ontrast.api.dtm\"\x9d\x07\n\x07Message\x12\x11\n\tclient_id\x18\x01 \x01(\t\x12\x19\n\rclient_number\x18\x02 \x01(\x05\x42\x02\x18\x01\x12\x18\n\x0c\x63lient_total\x18\x03 \x01(\x05\x42\x02\x18\x01\x12\x0b\n\x03pid\x18\x08 \x01(\x03\x12\x0c\n\x04ppid\x18\t \x01(\x03\x12\x15\n\rmessage_count\x18\x04 \x01(\x05\x12\x14\n\x0ctimestamp_ms\x18\x05 \x01(\x03\x12\x10\n\x08\x61pp_name\x18\x06 \x01(\t\x12\x14\n\x0c\x61pp_language\x18\x07 \x01(\t\x12\x10\n\x08\x61pp_path\x18\x1c \x01(\t\x12\x37\n\ragent_startup\x18\x0b \x01(\x0b\x32\x1e.contrast.api.dtm.AgentStartupH\x00\x12\x41\n\x12\x61pplication_create\x18\x0c \x01(\x0b\x32#.contrast.api.dtm.ApplicationCreateH\x00\x12;\n\x0fserver_activity\x18\n \x01(\x0b\x32 .contrast.api.dtm.ServerActivityH\x00\x12\x41\n\x12\x61pplication_update\x18\r \x01(\x0b\x32#.contrast.api.dtm.ApplicationUpdateH\x00\x12.\n\x08\x61\x63tivity\x18\x0e \x01(\x0b\x32\x1a.contrast.api.dtm.ActivityH\x00\x12\x39\n\x0eobserved_route\x18\x1a \x01(\x0b\x32\x1f.contrast.api.dtm.ObservedRouteH\x00\x12/\n\x07request\x18\x13 \x01(\x0b\x32\x1c.contrast.api.dtm.RawRequestH\x00\x12\x32\n\tprefilter\x18\x14 \x01(\x0b\x32\x1d.contrast.api.dtm.HttpRequestH\x00\x12\x34\n\npostfilter\x18\x16 \x01(\x0b\x32\x1e.contrast.api.dtm.HttpResponseH\x00\x12\x31\n\x08response\x18\x17 \x01(\x0b\x32\x1d.contrast.api.dtm.RawResponseH\x00\x12\x36\n\tsqli_sink\x18\x19 \x01(\x0b\x32!.contrast.api.dtm.SqliSinkRequestH\x00\x12&\n\x04poll\x18\x1b \x01(\x0b\x32\x16.contrast.api.dtm.PollH\x00\x12*\n\x04noop\x18\x18 \x01(\x0b\x32\x16.contrast.api.dtm.NoopB\x02\x18\x01H\x00\x42\x07\n\x05\x65vent\"\x06\n\x04Poll\"\x18\n\x04Noop\x12\x0c\n\x04\x66lag\x18\x01 \x01(\x05:\x02\x18\x01\"\x91\x03\n\x0eServerActivity\x12I\n\rlog_enhancers\x18\x10 \x03(\x0b\x32\x32.contrast.api.dtm.ServerActivity.LogEnhancersEntry\x12M\n\rip_blacklists\x18\x11 \x03(\x0b\x32\x32.contrast.api.dtm.ServerActivity.IpBlacklistsEntryB\x02\x18\x01\x12G\n\x0cip_denylists\x18\x12 \x03(\x0b\x32\x31.contrast.api.dtm.ServerActivity.IpDenylistsEntry\x1a\x33\n\x11LogEnhancersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x33\n\x11IpBlacklistsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x32\n\x10IpDenylistsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"\xa9\x02\n\x0c\x41gentStartup\x12\x13\n\x07version\x18\x02 \x01(\tB\x02\x18\x01\x12\x13\n\x0b\x65nvironment\x18\x03 \x01(\t\x12\x10\n\x04tags\x18\x04 \x01(\tB\x02\x18\x01\x12\x13\n\x0bserver_name\x18\x05 \x01(\t\x12\x13\n\x0bserver_path\x18\x06 \x01(\t\x12\x13\n\x0bserver_type\x18\x07 \x01(\t\x12\x16\n\x0eserver_version\x18\x08 \x01(\t\x12\x14\n\x0cheartbeat_ms\x18\t \x01(\x05\x12\x13\n\x0bserver_tags\x18\n \x01(\t\x12\x18\n\x10\x61pplication_tags\x18\x0b \x01(\t\x12\x14\n\x0clibrary_tags\x18\x0c \x01(\t\x12\x14\n\x0c\x66inding_tags\x18\r \x01(\t\x12\x15\n\ragent_version\x18\x0e \x01(\t\"\xcc\x01\n\x11\x41pplicationCreate\x12\r\n\x05group\x18\x01 \x01(\t\x12\x13\n\x0b\x61pp_version\x18\x02 \x01(\t\x12\x10\n\x04tags\x18\x03 \x01(\tB\x02\x18\x01\x12\x0c\n\x04\x63ode\x18\x04 \x01(\t\x12\x10\n\x08metadata\x18\x05 \x01(\t\x12\x12\n\nsession_id\x18\x06 \x01(\t\x12\x18\n\x10session_metadata\x18\x07 \x01(\t\x12\x33\n\x04mode\x18\n \x01(\x0b\x32%.contrast.api.dtm.InstrumentationMode\"6\n\x13InstrumentationMode\x12\x0f\n\x07protect\x18\x01 \x01(\x08\x12\x0e\n\x06\x61ssess\x18\x02 \x01(\x08\"\x81\x04\n\x11\x41pplicationUpdate\x12\x14\n\x0c\x63overage_loc\x18\x02 \x01(\x05\x12,\n\x08platform\x18\x04 \x01(\x0b\x32\x1a.contrast.api.dtm.Platform\x12$\n\x04view\x18\x05 \x01(\x0b\x32\x16.contrast.api.dtm.View\x12K\n\x0ctechnologies\x18\x06 \x03(\x0b\x32\x35.contrast.api.dtm.ApplicationUpdate.TechnologiesEntry\x12;\n\ncomponents\x18\x07 \x03(\x0b\x32\'.contrast.api.dtm.ArchitectureComponent\x12\x45\n\tlibraries\x18\x08 \x03(\x0b\x32\x32.contrast.api.dtm.ApplicationUpdate.LibrariesEntry\x12/\n\x06routes\x18\t \x03(\x0b\x32\x1f.contrast.api.dtm.RouteCoverage\x1a\x33\n\x11TechnologiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\x1aK\n\x0eLibrariesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.contrast.api.dtm.Library:\x02\x38\x01\"\x83\x07\n\x08\x41\x63tivity\x12\x33\n\x0chttp_request\x18\x01 \x01(\x0b\x32\x1d.contrast.api.dtm.HttpRequest\x12\x35\n\rhttp_response\x18\x02 \x01(\x0b\x32\x1e.contrast.api.dtm.HttpResponse\x12\x13\n\x0b\x64uration_ms\x18\x03 \x01(\x03\x12\x13\n\x0bquery_count\x18\x05 \x01(\x05\x12\x42\n\x0ctechnologies\x18\x07 \x03(\x0b\x32,.contrast.api.dtm.Activity.TechnologiesEntry\x12\x45\n\x0elibrary_usages\x18\x08 \x03(\x0b\x32-.contrast.api.dtm.Activity.LibraryUsagesEntry\x12>\n\rarchitectures\x18\t \x03(\x0b\x32\'.contrast.api.dtm.ArchitectureComponent\x12/\n\x06routes\x18\x13 \x03(\x0b\x32\x1f.contrast.api.dtm.RouteCoverage\x12/\n\x07results\x18\x0c \x03(\x0b\x32\x1e.contrast.api.dtm.AttackResult\x12+\n\x08\x66indings\x18\x0f \x03(\x0b\x32\x19.contrast.api.dtm.Finding\x12;\n\x08\x63ontrols\x18\x10 \x03(\x0b\x32).contrast.api.dtm.PossibleSecurityControl\x12\x18\n\x0c\x66inding_tags\x18\x11 \x01(\tB\x02\x18\x01\x12G\n\x0f\x64ynamic_sources\x18\x12 \x03(\x0b\x32..contrast.api.dtm.Activity.DynamicSourcesEntry\x1a\x33\n\x11TechnologiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\x1aZ\n\x12LibraryUsagesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32$.contrast.api.dtm.LibraryUsageUpdate:\x02\x38\x01\x1aV\n\x13\x44ynamicSourcesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.contrast.api.dtm.DynamicSource:\x02\x38\x01\"\xf0\x02\n\x07\x46inding\x12\x11\n\thash_code\x18\x01 \x01(\t\x12\x10\n\x08platform\x18\x02 \x01(\t\x12\x0f\n\x07rule_id\x18\x03 \x01(\t\x12\x10\n\x08\x65vidence\x18\x04 \x01(\t\x12=\n\nproperties\x18\x06 \x03(\x0b\x32).contrast.api.dtm.Finding.PropertiesEntry\x12,\n\x06\x65vents\x18\x07 \x03(\x0b\x32\x1c.contrast.api.dtm.TraceEvent\x12\x11\n\tpreflight\x18\x08 \x01(\t\x12\x10\n\x04tags\x18\t \x01(\tB\x02\x18\x01\x12\x0f\n\x07version\x18\n \x01(\x03\x12/\n\x06routes\x18\x0b \x03(\x0b\x32\x1f.contrast.api.dtm.RouteCoverage\x12\x16\n\nsession_id\x18\x0c \x01(\tB\x02\x18\x01\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x87\x02\n\rDynamicSource\x12\x12\n\nclass_name\x18\x01 \x01(\t\x12\x13\n\x0bmethod_name\x18\x02 \x01(\t\x12\x17\n\x0finstance_method\x18\x03 \x01(\x08\x12\x0e\n\x06target\x18\x04 \x01(\t\x12,\n\x06\x65vents\x18\x05 \x03(\x0b\x32\x1c.contrast.api.dtm.TraceEvent\x12\x43\n\nproperties\x18\x06 \x03(\x0b\x32/.contrast.api.dtm.DynamicSource.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"D\n\x17PossibleSecurityControl\x12\x0b\n\x03\x61pi\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x0e\n\x06traits\x18\x03 \x01(\x03\"I\n\x04View\x12\x12\n\naccessible\x18\x01 \x01(\x05\x12\x19\n\x11total_view_fields\x18\x02 \x01(\x05\x12\x12\n\nline_count\x18\x03 \x01(\x05\"^\n\x08Platform\x12\r\n\x05major\x18\x01 \x01(\t\x12\r\n\x05minor\x18\x02 \x01(\t\x12\r\n\x05\x62uild\x18\x03 \x01(\t\x12\x14\n\x0cservice_pack\x18\x04 \x01(\t\x12\x0f\n\x07updates\x18\x05 \x03(\t\"l\n\x15\x41rchitectureComponent\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0e\n\x06vendor\x18\x02 \x01(\t\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x13\n\x0bremote_host\x18\x04 \x01(\t\x12\x13\n\x0bremote_port\x18\x05 \x01(\x05\"\xd8\x01\n\x07Library\x12\x11\n\thash_code\x18\x01 \x01(\t\x12\x11\n\tfile_path\x18\x02 \x01(\t\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\t\x12\x0c\n\x04tloc\x18\x05 \x01(\x05\x12\x13\n\x0b\x63lass_count\x18\x06 \x01(\x05\x12\x18\n\x10used_class_count\x18\x07 \x01(\x05\x12\x13\n\x0binternal_ms\x18\x08 \x01(\x03\x12\x13\n\x0b\x65xternal_ms\x18\t \x01(\x03\x12\x10\n\x08manifest\x18\n \x01(\t\x12\x10\n\x04tags\x18\x0b \x01(\tB\x02\x18\x01\"\xb4\x01\n\x12LibraryUsageUpdate\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\x12\x11\n\thash_code\x18\x02 \x01(\t\x12I\n\x0b\x63lass_names\x18\x03 \x03(\x0b\x32\x34.contrast.api.dtm.LibraryUsageUpdate.ClassNamesEntry\x1a\x31\n\x0f\x43lassNamesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\"1\n\x07\x41\x64\x64ress\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0c\n\x04host\x18\x02 \x01(\t\x12\x0c\n\x04port\x18\x03 \x01(\x05\"\xea\x02\n\nRawRequest\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x14\n\x0ctimestamp_ms\x18\x02 \x01(\x03\x12\x14\n\x0crequest_line\x18\x04 \x01(\t\x12\x16\n\x0enormalized_uri\x18\x05 \x01(\t\x12\x11\n\tclient_ip\x18\x06 \x01(\t\x12\x19\n\x11\x63lient_ip_version\x18\x07 \x01(\x05\x12\x13\n\x0b\x63lient_port\x18\x08 \x01(\x05\x12\x11\n\tserver_ip\x18\n \x01(\t\x12\x19\n\x11server_ip_version\x18\x0b \x01(\x05\x12\x13\n\x0bserver_port\x18\x0c \x01(\x05\x12\x35\n\x0frequest_headers\x18\x0e \x03(\x0b\x32\x1c.contrast.api.dtm.SimplePair\x12\x37\n\x11multipart_headers\x18\x11 \x03(\x0b\x32\x1c.contrast.api.dtm.SimplePair\x12\x14\n\x0crequest_body\x18\x10 \x01(\x0c\"\x97\x01\n\x0bRawResponse\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x14\n\x0ctimestamp_ms\x18\x02 \x01(\x03\x12\x15\n\rresponse_code\x18\x03 \x01(\x05\x12\x36\n\x10response_headers\x18\x0e \x03(\x0b\x32\x1c.contrast.api.dtm.SimplePair\x12\x15\n\rresponse_body\x18\x10 \x01(\x0c\"\xd7\x0b\n\x0bHttpRequest\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x14\n\x0ctimestamp_ms\x18\x02 \x01(\x03\x12)\n\x06sender\x18\x03 \x01(\x0b\x32\x19.contrast.api.dtm.Address\x12+\n\x08receiver\x18\x04 \x01(\x0b\x32\x19.contrast.api.dtm.Address\x12\x10\n\x08protocol\x18\x05 \x01(\t\x12\x0f\n\x07version\x18\x06 \x01(\t\x12\x0e\n\x06method\x18\x07 \x01(\t\x12\x0b\n\x03raw\x18\x08 \x01(\t\x12\x19\n\x11parsed_connection\x18\t \x01(\x08\x12\x0b\n\x03uri\x18\n \x01(\t\x12\x16\n\x0enormalized_uri\x18\x0b \x01(\t\x12\x0f\n\x07\x63ontext\x18\x0c \x01(\t\x12\x0c\n\x04path\x18\r \x01(\t\x12\x14\n\x0cquery_string\x18\x0e \x01(\t\x12]\n\x19normalized_request_params\x18\x10 \x03(\x0b\x32:.contrast.api.dtm.HttpRequest.NormalizedRequestParamsEntry\x12\x37\n\x11multipart_headers\x18\x33 \x03(\x0b\x32\x1c.contrast.api.dtm.SimplePair\x12J\n\x0frequest_headers\x18\x14 \x03(\x0b\x32\x31.contrast.api.dtm.HttpRequest.RequestHeadersEntry\x12\x1e\n\x16parsed_request_headers\x18\x15 \x01(\x08\x12\x63\n\x1anormalized_request_headers\x18\x16 \x03(\x0b\x32;.contrast.api.dtm.HttpRequest.NormalizedRequestHeadersEntryB\x02\x18\x01\x12P\n\x12normalized_cookies\x18\x17 \x03(\x0b\x32\x34.contrast.api.dtm.HttpRequest.NormalizedCookiesEntry\x12\x18\n\x0crequest_body\x18\x1e \x01(\tB\x02\x18\x01\x12\x1b\n\x13parsed_request_body\x18\x1f \x01(\x08\x12\x41\n\rdocument_type\x18  \x01(\x0e\x32*.contrast.api.dtm.HttpRequest.DocumentType\x12\x1b\n\x13request_body_binary\x18! \x01(\x0c\x12\x44\n\x0e\x61uthentication\x18( \x01(\x0e\x32,.contrast.api.dtm.HttpRequest.Authentication\x12\x11\n\tprincipal\x18) \x01(\t\x12\x13\n\x0bserver_info\x18\x32 \x01(\t\x1aV\n\x1cNormalizedRequestParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.contrast.api.dtm.Pair:\x02\x38\x01\x1a\x35\n\x13RequestHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1aW\n\x1dNormalizedRequestHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.contrast.api.dtm.Pair:\x02\x38\x01\x1aP\n\x16NormalizedCookiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.contrast.api.dtm.Pair:\x02\x38\x01\"n\n\x0e\x41uthentication\x12\x13\n\x0fUNAUTHENTICATED\x10\x00\x12\x11\n\rAUTHENTICATED\x10\x01\x12\x0e\n\nLOGIN_FAIL\x10\x02\x12\x11\n\rLOGIN_SUCCESS\x10\x03\x12\x11\n\rLOGIN_UNKNOWN\x10\x04\"-\n\x0c\x44ocumentType\x12\n\n\x06NORMAL\x10\x00\x12\x08\n\x04JSON\x10\x01\x12\x07\n\x03XML\x10\x02\"e\n\tRpcStatus\x12.\n\x04\x63ode\x18\x01 \x01(\x0e\x32 .contrast.api.dtm.RpcStatus.Code\x12\x0b\n\x03msg\x18\x02 \x01(\t\"\x1b\n\x04\x43ode\x12\x06\n\x02OK\x10\x00\x12\x0b\n\x07UNKNOWN\x10\x01\"\x9d\x01\n\x0fSqliSinkRequest\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12:\n\x07\x64ialect\x18\x02 \x01(\x0e\x32).contrast.api.dtm.SqliSinkRequest.Dialect\x12\r\n\x05query\x18\x03 \x01(\t\"1\n\x07\x44ialect\x12\t\n\x05MYSQL\x10\x00\x12\x0e\n\nPOSTGRESQL\x10\x01\x12\x0b\n\x07UNKNOWN\x10\x63\"\x82\x02\n\x10SqliSinkResponse\x12+\n\x06status\x18\x01 \x01(\x0b\x32\x1b.contrast.api.dtm.RpcStatus\x12/\n\nuser_input\x18\x03 \x01(\x0b\x32\x1b.contrast.api.dtm.UserInput\x12\x35\n\x04sqli\x18\n \x01(\x0b\x32%.contrast.api.dtm.SqlInjectionDetailsH\x00\x12N\n\rsqli_semantic\x18\x0b \x01(\x0b\x32\x35.contrast.api.dtm.SqlInjectionSemanticAnalysisDetailsH\x00\x42\t\n\x07\x66inding\"\xd1\x04\n\x0cHttpResponse\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x14\n\x0ctimestamp_ms\x18\x02 \x01(\x03\x12\x15\n\rresponse_code\x18\x03 \x01(\x05\x12Q\n\x10response_headers\x18\x04 \x03(\x0b\x32\x33.contrast.api.dtm.HttpResponse.ResponseHeadersEntryB\x02\x18\x01\x12#\n\x17parsed_response_headers\x18\x05 \x01(\x08\x42\x02\x18\x01\x12\x62\n\x1bnormalized_response_headers\x18\x06 \x03(\x0b\x32=.contrast.api.dtm.HttpResponse.NormalizedResponseHeadersEntry\x12\x19\n\rresponse_body\x18\n \x01(\tB\x02\x18\x01\x12\x1c\n\x14parsed_response_body\x18\x0b \x01(\x08\x12\x41\n\rdocument_type\x18\x0c \x01(\x0e\x32*.contrast.api.dtm.HttpRequest.DocumentType\x12\x1c\n\x14response_body_binary\x18\r \x01(\x0c\x1a\x36\n\x14ResponseHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1aX\n\x1eNormalizedResponseHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.contrast.api.dtm.Pair:\x02\x38\x01\"(\n\nSimplePair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"#\n\x04Pair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0e\n\x06values\x18\x02 \x03(\t\"\x8c\x03\n\x0c\x41ttackResult\x12=\n\x08response\x18\x01 \x01(\x0e\x32+.contrast.api.dtm.AttackResult.ResponseType\x12\x0f\n\x07rule_id\x18\x02 \x01(\t\x12\x31\n\x07samples\x18\x06 \x03(\x0b\x32 .contrast.api.dtm.RaspRuleSample\x12\x10\n\x04tags\x18\x07 \x01(\tB\x02\x18\x01\x12\x16\n\x0e\x61\x63\x63_updated_ms\x18\x10 \x01(\x03\x12\x11\n\tacc_limit\x18\x11 \x01(\x05\x12\x15\n\racc_window_ms\x18\x12 \x01(\x03\x12\x11\n\tacc_count\x18\x13 \x01(\x05\x12\x10\n\x08\x61\x63\x63_mode\x18\x15 \x01(\t\x12\x0f\n\x07\x61\x63\x63_msg\x18\x16 \x01(\t\"o\n\x0cResponseType\x12\r\n\tNO_ACTION\x10\x00\x12\x0b\n\x07\x42LOCKED\x10\x01\x12\r\n\tMONITORED\x10\x02\x12\n\n\x06PROBED\x10\x03\x12\x18\n\x14\x42LOCKED_AT_PERIMETER\x10\x04\x12\x0e\n\nAGGREGATED\x10\x05\"\x87\x10\n\x0eRaspRuleSample\x12\x14\n\x0ctimestamp_ms\x18\x02 \x01(\x03\x12\x41\n\x14stack_trace_elements\x18\x04 \x03(\x0b\x32#.contrast.api.dtm.StackTraceElement\x12/\n\nuser_input\x18\x05 \x01(\x0b\x32\x1b.contrast.api.dtm.UserInput\x12:\n\x0b\x62rute_force\x18\x10 \x01(\x0b\x32#.contrast.api.dtm.BruteForceDetailsH\x00\x12:\n\x0b\x62ot_blocker\x18\x11 \x01(\x0b\x32#.contrast.api.dtm.BotBlockerDetailsH\x00\x12\x35\n\x04\x63mdi\x18\x12 \x01(\x0b\x32%.contrast.api.dtm.CmdInjectionDetailsH\x00\x12-\n\x04\x63srf\x18\x13 \x01(\x0b\x32\x1d.contrast.api.dtm.CsrfDetailsH\x00\x12/\n\x03\x63ve\x18\x14 \x01(\x0b\x32\x1c.contrast.api.dtm.CveDetailsB\x02\x18\x01H\x00\x12V\n\x19untrusted_deserialization\x18\x15 \x01(\x0b\x32\x31.contrast.api.dtm.UntrustedDeserializationDetailsH\x00\x12<\n\x0c\x65l_injection\x18\x16 \x01(\x0b\x32$.contrast.api.dtm.ElInjectionDetailsH\x00\x12@\n\x0cip_blacklist\x18\x17 \x01(\x0b\x32$.contrast.api.dtm.IpBlacklistDetailsB\x02\x18\x01H\x00\x12:\n\x0bip_denylist\x18- \x01(\x0b\x32#.contrast.api.dtm.IpDenylistDetailsH\x00\x12\x44\n\x11mark_of_the_beast\x18\x18 \x01(\x0b\x32\'.contrast.api.dtm.MarkOfTheBeastDetailsH\x00\x12@\n\x0epadding_oracle\x18\x19 \x01(\x0b\x32&.contrast.api.dtm.PaddingOracleDetailsH\x00\x12@\n\x0epath_traversal\x18\x1a \x01(\x0b\x32&.contrast.api.dtm.PathTraversalDetailsH\x00\x12\x30\n\x06re_dos\x18\x1b \x01(\x0b\x32\x1e.contrast.api.dtm.ReDosDetailsH\x00\x12\x35\n\x04sqli\x18\x1c \x01(\x0b\x32%.contrast.api.dtm.SqlInjectionDetailsH\x00\x12-\n\x04ssrf\x18\x1d \x01(\x0b\x32\x1d.contrast.api.dtm.SsrfDetailsH\x00\x12>\n\rvirtual_patch\x18\x1e \x01(\x0b\x32%.contrast.api.dtm.VirtualPatchDetailsH\x00\x12+\n\x03xss\x18\x1f \x01(\x0b\x32\x1c.contrast.api.dtm.XssDetailsH\x00\x12+\n\x03xxe\x18  \x01(\x0b\x32\x1c.contrast.api.dtm.XxeDetailsH\x00\x12:\n\x07no_sqli\x18! \x01(\x0b\x32\'.contrast.api.dtm.NoSqlInjectionDetailsH\x00\x12H\n\x10method_tampering\x18\" \x01(\x0b\x32,.contrast.api.dtm.HttpMethodTamperingDetailsH\x00\x12\x41\n\rcve_2017_0898\x18# \x01(\x0b\x32$.contrast.api.dtm.Cve20170898DetailsB\x02\x18\x01H\x00\x12\x43\n\x0e\x63ve_2017_14033\x18$ \x01(\x0b\x32%.contrast.api.dtm.Cve201714033DetailsB\x02\x18\x01H\x00\x12\x43\n\x0e\x63ve_2017_14064\x18% \x01(\x0b\x32%.contrast.api.dtm.Cve201714064DetailsB\x02\x18\x01H\x00\x12\x43\n\x0e\x63ve_2017_17405\x18& \x01(\x0b\x32%.contrast.api.dtm.Cve201717405DetailsB\x02\x18\x01H\x00\x12<\n\x0cmod_security\x18\' \x01(\x0b\x32$.contrast.api.dtm.ModSecurityDetailsH\x00\x12\x38\n\x08\x65ndpoint\x18( \x01(\x0b\x32$.contrast.api.dtm.EndpointBanDetailsH\x00\x12N\n\rsqli_semantic\x18) \x01(\x0b\x32\x35.contrast.api.dtm.SqlInjectionSemanticAnalysisDetailsH\x00\x12N\n\rcmdi_semantic\x18* \x01(\x0b\x32\x35.contrast.api.dtm.CmdInjectionSemanticAnalysisDetailsH\x00\x12Y\n\x17path_traversal_semantic\x18+ \x01(\x0b\x32\x36.contrast.api.dtm.PathTraversalSemanticAnalysisDetailsH\x00\x12\x36\n\x04ssjs\x18, \x01(\x0b\x32&.contrast.api.dtm.SsjsInjectionDetailsH\x00\x42\t\n\x07\x64\x65tails\"L\n\x12\x45ndpointBanDetails\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x13\n\x0brequest_uri\x18\x02 \x01(\t\x12\x15\n\rresponse_code\x18\x03 \x01(\x05\"l\n\x12ModSecurityDetails\x12\x16\n\x0etransaction_id\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\x05\x12.\n\x04logs\x18\x03 \x03(\x0b\x32 .contrast.api.dtm.ModSecurityLog\"\x8e\x01\n\x0eModSecurityLog\x12\x0f\n\x07message\x18\x02 \x01(\t\x12<\n\x06params\x18\x03 \x03(\x0b\x32,.contrast.api.dtm.ModSecurityLog.ParamsEntry\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\"\n\x11\x42ruteForceDetails\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\"4\n\x11\x42otBlockerDetails\x12\x0b\n\x03\x62ot\x18\x01 \x01(\t\x12\x12\n\nuser_agent\x18\x02 \x01(\t\"J\n\x13\x43mdInjectionDetails\x12\x0f\n\x07\x63ommand\x18\x01 \x01(\t\x12\x11\n\tstart_idx\x18\x02 \x01(\x05\x12\x0f\n\x07\x65nd_idx\x18\x03 \x01(\x05\"\xc2\x01\n#CmdInjectionSemanticAnalysisDetails\x12\x0f\n\x07\x63ommand\x18\x01 \x01(\t\x12O\n\x08\x66indings\x18\x02 \x03(\x0e\x32=.contrast.api.dtm.CmdInjectionSemanticAnalysisDetails.Finding\"9\n\x07\x46inding\x12\x0c\n\x08\x43HAINING\x10\x00\x12\x0c\n\x08\x42\x41\x43KDOOR\x10\x01\x12\x12\n\x0e\x44\x41NGEROUS_PATH\x10\x02\"@\n\x0b\x43srfDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x65xpected\x18\x02 \x01(\t\x12\x11\n\tpresented\x18\x03 \x01(\t\"G\n\nCveDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\x0b\n\x03\x63ve\x18\x03 \x01(\t\x12\x0f\n\x07library\x18\x04 \x01(\t\"\xa8\x01\n\x12\x43ve20170898Details\x12\x15\n\rformat_string\x18\x01 \x01(\t\x12\x1e\n\x16malicious_format_group\x18\x02 \x01(\t\x12\x1a\n\x12precision_argument\x18\x03 \x01(\t\x12\x16\n\x0evalue_argument\x18\x04 \x01(\t\x12\x11\n\targuments\x18\x05 \x01(\t\x12\x14\n\x0cruby_version\x18\x06 \x01(\t\"9\n\x13\x43ve201714033Details\x12\x0c\n\x04\x61sn1\x18\x01 \x01(\t\x12\x14\n\x0cruby_version\x18\x02 \x01(\t\"P\n\x13\x43ve201714064Details\x12\r\n\x05space\x18\x01 \x01(\t\x12\x14\n\x0cruby_version\x18\x02 \x01(\t\x12\x14\n\x0cjson_version\x18\x03 \x01(\t\"N\n\x13\x43ve201717405Details\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x11\n\tfile_name\x18\x02 \x01(\t\x12\x14\n\x0cruby_version\x18\x03 \x01(\t\"C\n\x1aHttpMethodTamperingDetails\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x15\n\rresponse_code\x18\x02 \x01(\x05\"H\n\x1fUntrustedDeserializationDetails\x12\x0f\n\x07\x63ommand\x18\x01 \x01(\x08\x12\x14\n\x0c\x64\x65serializer\x18\x02 \x01(\t\"S\n\x12\x45lInjectionDetails\x12\x12\n\nattr_value\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x18\n\x10library_filename\x18\x03 \x01(\t\"2\n\x12IpBlacklistDetails\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0c\n\x04uuid\x18\x02 \x01(\t:\x02\x18\x01\"-\n\x11IpDenylistDetails\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0c\n\x04uuid\x18\x02 \x01(\t\"H\n\x15MarkOfTheBeastDetails\x12\x0e\n\x06number\x18\x01 \x01(\t\x12\x0e\n\x06vendor\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\"E\n\x14PaddingOracleDetails\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0e\n\x06period\x18\x02 \x01(\x05\x12\x0c\n\x04hits\x18\x03 \x01(\x05\"$\n\x14PathTraversalDetails\x12\x0c\n\x04path\x18\x01 \x01(\t\"\xd3\x01\n$PathTraversalSemanticAnalysisDetails\x12\x0c\n\x04path\x18\x01 \x01(\t\x12P\n\x08\x66indings\x18\x02 \x03(\x0e\x32>.contrast.api.dtm.PathTraversalSemanticAnalysisDetails.Finding\"K\n\x07\x46inding\x12&\n\"CUSTOM_CODE_ACCESSING_SYSTEM_FILES\x10\x00\x12\x18\n\x14\x43OMMON_FILE_EXPLOITS\x10\x01\"l\n\x0cReDosDetails\x12\x16\n\x0etimes_accessed\x18\x01 \x01(\x05\x12\x19\n\x11\x65xecution_time_ms\x18\x02 \x01(\x03\x12\x0f\n\x07pattern\x18\x03 \x01(\t\x12\x18\n\x10offending_string\x18\x04 \x01(\t\"m\n\x14SsjsInjectionDetails\x12\x13\n\x0b\x63ode_string\x18\x01 \x01(\t\x12\x11\n\tstart_idx\x18\x02 \x01(\x05\x12\x0f\n\x07\x65nd_idx\x18\x03 \x01(\x05\x12\x1c\n\x14\x62oundary_overrun_idx\x18\x04 \x01(\x05\"\x82\x01\n\x13SqlInjectionDetails\x12\x11\n\tstart_idx\x18\x01 \x01(\x05\x12\x0f\n\x07\x65nd_idx\x18\x02 \x01(\x05\x12\x1c\n\x14\x62oundary_overrun_idx\x18\x03 \x01(\x05\x12\x1a\n\x12input_boundary_idx\x18\x04 \x01(\x05\x12\r\n\x05query\x18\x05 \x01(\t\"\xda\x01\n#SqlInjectionSemanticAnalysisDetails\x12\r\n\x05query\x18\x01 \x01(\t\x12N\n\x07\x66inding\x18\x02 \x01(\x0e\x32=.contrast.api.dtm.SqlInjectionSemanticAnalysisDetails.Finding\"T\n\x07\x46inding\x12\x0c\n\x08\x43HAINING\x10\x00\x12\x14\n\x10SUSPICIOUS_UNION\x10\x01\x12\x16\n\x12\x44\x41NGEROUS_FUNCTION\x10\x02\x12\r\n\tTAUTOLOGY\x10\x03\"\x84\x01\n\x15NoSqlInjectionDetails\x12\x11\n\tstart_idx\x18\x01 \x01(\x05\x12\x0f\n\x07\x65nd_idx\x18\x02 \x01(\x05\x12\x1c\n\x14\x62oundary_overrun_idx\x18\x03 \x01(\x05\x12\x1a\n\x12input_boundary_idx\x18\x04 \x01(\x05\x12\r\n\x05query\x18\x05 \x01(\t\"\x1a\n\x0bSsrfDetails\x12\x0b\n\x03url\x18\x01 \x01(\t\"#\n\x13VirtualPatchDetails\x12\x0c\n\x04uuid\x18\x01 \x01(\t\"H\n\nXssDetails\x12\r\n\x05input\x18\x01 \x01(\t\x12+\n\x07matches\x18\x02 \x03(\x0b\x32\x1a.contrast.api.dtm.XssMatch\"G\n\x08XssMatch\x12\x19\n\x11\x65vidence_start_ms\x18\x01 \x01(\x03\x12\x10\n\x08\x65vidence\x18\x02 \x01(\t\x12\x0e\n\x06offset\x18\x03 \x01(\x05\"\x89\x01\n\nXxeDetails\x12\x0b\n\x03xml\x18\x01 \x01(\t\x12\x35\n\x11\x64\x65\x63lared_entities\x18\x02 \x03(\x0b\x32\x1a.contrast.api.dtm.XxeMatch\x12\x37\n\x11\x65ntities_resolved\x18\x03 \x03(\x0b\x32\x1c.contrast.api.dtm.XxeWrapper\".\n\x08XxeMatch\x12\x11\n\tstart_idx\x18\x01 \x01(\x05\x12\x0f\n\x07\x65nd_idx\x18\x02 \x01(\x05\"2\n\nXxeWrapper\x12\x11\n\tsystem_id\x18\x01 \x01(\t\x12\x11\n\tpublic_id\x18\x02 \x01(\t\"\xd0\x04\n\tUserInput\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\t\x12\x39\n\ninput_type\x18\x05 \x01(\x0e\x32%.contrast.api.dtm.UserInput.InputType\x12\x41\n\rdocument_type\x18\x06 \x01(\x0e\x32*.contrast.api.dtm.HttpRequest.DocumentType\x12\x13\n\x0bmatcher_ids\x18\x07 \x03(\t\"\x85\x03\n\tInputType\x12\x12\n\x0eUNDEFINED_TYPE\x10\x00\x12\x08\n\x04\x42ODY\x10\x01\x12\x0f\n\x0b\x43OOKIE_NAME\x10\x02\x12\x10\n\x0c\x43OOKIE_VALUE\x10\x03\x12\n\n\x06HEADER\x10\x04\x12\x12\n\x0ePARAMETER_NAME\x10\x05\x12\x13\n\x0fPARAMETER_VALUE\x10\x06\x12\x0f\n\x0bQUERYSTRING\x10\x07\x12\x07\n\x03URI\x10\x08\x12\n\n\x06SOCKET\x10\t\x12\x0e\n\nJSON_VALUE\x10\n\x12\x16\n\x12JSON_ARRAYED_VALUE\x10\x0b\x12\x1a\n\x16MULTIPART_CONTENT_TYPE\x10\x10\x12\x13\n\x0fMULTIPART_VALUE\x10\x11\x12\x18\n\x14MULTIPART_FIELD_NAME\x10\x12\x12\x12\n\x0eMULTIPART_NAME\x10\x13\x12\r\n\tXML_VALUE\x10\x14\x12\r\n\tDWR_VALUE\x10\x15\x12\n\n\x06METHOD\x10\x16\x12\x0b\n\x07REQUEST\x10\x17\x12\x11\n\rURL_PARAMETER\x10\x18\x12\x0b\n\x07UNKNOWN\x10\x63\"i\n\x11StackTraceElement\x12\x17\n\x0f\x64\x65\x63laring_class\x18\x01 \x01(\t\x12\x13\n\x0bmethod_name\x18\x02 \x01(\t\x12\x11\n\tfile_name\x18\x03 \x01(\t\x12\x13\n\x0bline_number\x18\x04 \x01(\x05\"\xa0\x07\n\nTraceEvent\x12\x33\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32#.contrast.api.dtm.TraceEvent.Action\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.contrast.api.dtm.TraceEvent.EventType\x12\x14\n\x0ctimestamp_ms\x18\x03 \x01(\x03\x12\x0e\n\x06thread\x18\x04 \x01(\t\x12\x38\n\tsignature\x18\x06 \x01(\x0b\x32%.contrast.api.dtm.TraceEventSignature\x12\x12\n\nfield_name\x18\x07 \x01(\t\x12\x0f\n\x07\x63ontext\x18\x08 \x01(\t\x12\x0c\n\x04\x63ode\x18\t \x01(\t\x12\x32\n\x06object\x18\x10 \x01(\x0b\x32\".contrast.api.dtm.TraceEventObject\x12/\n\x03ret\x18\x11 \x01(\x0b\x32\".contrast.api.dtm.TraceEventObject\x12\x30\n\x04\x61rgs\x18\x12 \x03(\x0b\x32\".contrast.api.dtm.TraceEventObject\x12+\n\x05stack\x18\x13 \x03(\x0b\x32\x1c.contrast.api.dtm.TraceStack\x12\x39\n\revent_sources\x18\x14 \x03(\x0b\x32\".contrast.api.dtm.TraceEventSource\x12\x0e\n\x06source\x18\x15 \x01(\t\x12\x0e\n\x06target\x18\x16 \x01(\t\x12\x37\n\x0ctaint_ranges\x18\x17 \x03(\x0b\x32!.contrast.api.dtm.TraceTaintRange\x12\x11\n\tobject_id\x18\x18 \x01(\x03\x12;\n\x11parent_object_ids\x18\x19 \x03(\x0b\x32 .contrast.api.dtm.ParentObjectId\"\xa9\x01\n\x06\x41\x63tion\x12\x0c\n\x08\x43REATION\x10\x00\x12\x0f\n\x0bPROPAGATION\x10\x01\x12\x0b\n\x07TRIGGER\x10\x02\x12\x07\n\x03TAG\x10\x03\x12\x07\n\x03\x41\x32\x41\x10\x04\x12\x07\n\x03\x41\x32P\x10\x05\x12\x07\n\x03\x41\x32O\x10\x06\x12\x07\n\x03\x41\x32R\x10\x07\x12\x07\n\x03O2A\x10\x08\x12\x07\n\x03O2P\x10\t\x12\x07\n\x03O2O\x10\n\x12\x07\n\x03O2R\x10\x0b\x12\x07\n\x03P2A\x10\x0c\x12\x07\n\x03P2P\x10\r\x12\x07\n\x03P2O\x10\x0e\x12\x07\n\x03P2R\x10\x0f\"@\n\tEventType\x12\x0f\n\x0bTYPE_METHOD\x10\x00\x12\x14\n\x10TYPE_PROPAGATION\x10\x01\x12\x0c\n\x08TYPE_TAG\x10\x02\"\x9f\x01\n\x13TraceEventSignature\x12\x13\n\x0breturn_type\x18\x01 \x01(\t\x12\x12\n\nclass_name\x18\x02 \x01(\t\x12\x13\n\x0bmethod_name\x18\x03 \x01(\t\x12\x11\n\targ_types\x18\x04 \x03(\t\x12\x13\n\x0b\x63onstructor\x18\x05 \x01(\x08\x12\x13\n\x0bvoid_method\x18\x06 \x01(\x08\x12\r\n\x05\x66lags\x18\x07 \x01(\x05\".\n\x10TraceEventSource\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"e\n\x10TraceEventObject\x12\x0f\n\x07tracked\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\t\x12\x31\n\x06ranges\x18\x03 \x03(\x0b\x32!.contrast.api.dtm.TraceTaintRange\"-\n\x0fTraceTaintRange\x12\x0b\n\x03tag\x18\x01 \x01(\t\x12\r\n\x05range\x18\x02 \x01(\t\"\x1c\n\x0eParentObjectId\x12\n\n\x02id\x18\x01 \x01(\x03\"\x91\x01\n\nTraceStack\x12\x11\n\tsignature\x18\x01 \x01(\t\x12\x17\n\x0f\x64\x65\x63laring_class\x18\x02 \x01(\t\x12\x13\n\x0bmethod_name\x18\x03 \x01(\t\x12\x11\n\tfile_name\x18\x04 \x01(\t\x12\x13\n\x0bline_number\x18\x05 \x01(\x05\x12\x0c\n\x04type\x18\x06 \x01(\t\x12\x0c\n\x04\x65val\x18\x07 \x01(\t\"H\n\rRouteCoverage\x12\x0c\n\x04verb\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\r\n\x05route\x18\x03 \x01(\t\x12\r\n\x05\x63ount\x18\x04 \x01(\x05\"\x8a\x01\n\rObservedRoute\x12\x11\n\tsignature\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\x0c\n\x04verb\x18\x03 \x01(\t\x12\x16\n\nsession_id\x18\x04 \x01(\tB\x02\x18\x01\x12\x33\n\x07sources\x18\x05 \x03(\x0b\x32\".contrast.api.dtm.TraceEventSource\"\x11\n\x0f\x43mdiSinkRequest\"\x12\n\x10\x43mdiSinkResponse\"\x0f\n\rPtSinkRequest\"\x10\n\x0ePtSinkResponseB<Z:bitbucket.org/contrastsecurity/go-speedracer-go/from_agentb\x06proto3'
)



_HTTPREQUEST_AUTHENTICATION = _descriptor.EnumDescriptor(
  name='Authentication',
  full_name='contrast.api.dtm.HttpRequest.Authentication',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNAUTHENTICATED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AUTHENTICATED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOGIN_FAIL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOGIN_SUCCESS', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOGIN_UNKNOWN', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=6677,
  serialized_end=6787,
)
_sym_db.RegisterEnumDescriptor(_HTTPREQUEST_AUTHENTICATION)

_HTTPREQUEST_DOCUMENTTYPE = _descriptor.EnumDescriptor(
  name='DocumentType',
  full_name='contrast.api.dtm.HttpRequest.DocumentType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NORMAL', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JSON', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='XML', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=6789,
  serialized_end=6834,
)
_sym_db.RegisterEnumDescriptor(_HTTPREQUEST_DOCUMENTTYPE)

_RPCSTATUS_CODE = _descriptor.EnumDescriptor(
  name='Code',
  full_name='contrast.api.dtm.RpcStatus.Code',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=6910,
  serialized_end=6937,
)
_sym_db.RegisterEnumDescriptor(_RPCSTATUS_CODE)

_SQLISINKREQUEST_DIALECT = _descriptor.EnumDescriptor(
  name='Dialect',
  full_name='contrast.api.dtm.SqliSinkRequest.Dialect',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MYSQL', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POSTGRESQL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=2, number=99,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=7048,
  serialized_end=7097,
)
_sym_db.RegisterEnumDescriptor(_SQLISINKREQUEST_DIALECT)

_ATTACKRESULT_RESPONSETYPE = _descriptor.EnumDescriptor(
  name='ResponseType',
  full_name='contrast.api.dtm.AttackResult.ResponseType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NO_ACTION', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BLOCKED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MONITORED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROBED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BLOCKED_AT_PERIMETER', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AGGREGATED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=8321,
  serialized_end=8432,
)
_sym_db.RegisterEnumDescriptor(_ATTACKRESULT_RESPONSETYPE)

_CMDINJECTIONSEMANTICANALYSISDETAILS_FINDING = _descriptor.EnumDescriptor(
  name='Finding',
  full_name='contrast.api.dtm.CmdInjectionSemanticAnalysisDetails.Finding',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CHAINING', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BACKDOOR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DANGEROUS_PATH', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11129,
  serialized_end=11186,
)
_sym_db.RegisterEnumDescriptor(_CMDINJECTIONSEMANTICANALYSISDETAILS_FINDING)

_PATHTRAVERSALSEMANTICANALYSISDETAILS_FINDING = _descriptor.EnumDescriptor(
  name='Finding',
  full_name='contrast.api.dtm.PathTraversalSemanticAnalysisDetails.Finding',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CUSTOM_CODE_ACCESSING_SYSTEM_FILES', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMMON_FILE_EXPLOITS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=12366,
  serialized_end=12441,
)
_sym_db.RegisterEnumDescriptor(_PATHTRAVERSALSEMANTICANALYSISDETAILS_FINDING)

_SQLINJECTIONSEMANTICANALYSISDETAILS_FINDING = _descriptor.EnumDescriptor(
  name='Finding',
  full_name='contrast.api.dtm.SqlInjectionSemanticAnalysisDetails.Finding',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CHAINING', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUSPICIOUS_UNION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DANGEROUS_FUNCTION', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TAUTOLOGY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=12932,
  serialized_end=13016,
)
_sym_db.RegisterEnumDescriptor(_SQLINJECTIONSEMANTICANALYSISDETAILS_FINDING)

_USERINPUT_INPUTTYPE = _descriptor.EnumDescriptor(
  name='InputType',
  full_name='contrast.api.dtm.UserInput.InputType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDEFINED_TYPE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BODY', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COOKIE_NAME', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COOKIE_VALUE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HEADER', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PARAMETER_NAME', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PARAMETER_VALUE', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='QUERYSTRING', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='URI', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SOCKET', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JSON_VALUE', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JSON_ARRAYED_VALUE', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MULTIPART_CONTENT_TYPE', index=12, number=16,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MULTIPART_VALUE', index=13, number=17,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MULTIPART_FIELD_NAME', index=14, number=18,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MULTIPART_NAME', index=15, number=19,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='XML_VALUE', index=16, number=20,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DWR_VALUE', index=17, number=21,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='METHOD', index=18, number=22,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REQUEST', index=19, number=23,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='URL_PARAMETER', index=20, number=24,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=21, number=99,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=13809,
  serialized_end=14198,
)
_sym_db.RegisterEnumDescriptor(_USERINPUT_INPUTTYPE)

_TRACEEVENT_ACTION = _descriptor.EnumDescriptor(
  name='Action',
  full_name='contrast.api.dtm.TraceEvent.Action',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CREATION', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROPAGATION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRIGGER', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TAG', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='A2A', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='A2P', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='A2O', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='A2R', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='O2A', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='O2P', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='O2O', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='O2R', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='P2A', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='P2P', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='P2O', index=14, number=14,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='P2R', index=15, number=15,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15001,
  serialized_end=15170,
)
_sym_db.RegisterEnumDescriptor(_TRACEEVENT_ACTION)

_TRACEEVENT_EVENTTYPE = _descriptor.EnumDescriptor(
  name='EventType',
  full_name='contrast.api.dtm.TraceEvent.EventType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_METHOD', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_PROPAGATION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_TAG', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15172,
  serialized_end=15236,
)
_sym_db.RegisterEnumDescriptor(_TRACEEVENT_EVENTTYPE)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='contrast.api.dtm.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_id', full_name='contrast.api.dtm.Message.client_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_number', full_name='contrast.api.dtm.Message.client_number', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_total', full_name='contrast.api.dtm.Message.client_total', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pid', full_name='contrast.api.dtm.Message.pid', index=3,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ppid', full_name='contrast.api.dtm.Message.ppid', index=4,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_count', full_name='contrast.api.dtm.Message.message_count', index=5,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_ms', full_name='contrast.api.dtm.Message.timestamp_ms', index=6,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_name', full_name='contrast.api.dtm.Message.app_name', index=7,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_language', full_name='contrast.api.dtm.Message.app_language', index=8,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_path', full_name='contrast.api.dtm.Message.app_path', index=9,
      number=28, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='agent_startup', full_name='contrast.api.dtm.Message.agent_startup', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='application_create', full_name='contrast.api.dtm.Message.application_create', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_activity', full_name='contrast.api.dtm.Message.server_activity', index=12,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='application_update', full_name='contrast.api.dtm.Message.application_update', index=13,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='activity', full_name='contrast.api.dtm.Message.activity', index=14,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='observed_route', full_name='contrast.api.dtm.Message.observed_route', index=15,
      number=26, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request', full_name='contrast.api.dtm.Message.request', index=16,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prefilter', full_name='contrast.api.dtm.Message.prefilter', index=17,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='postfilter', full_name='contrast.api.dtm.Message.postfilter', index=18,
      number=22, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response', full_name='contrast.api.dtm.Message.response', index=19,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sqli_sink', full_name='contrast.api.dtm.Message.sqli_sink', index=20,
      number=25, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='poll', full_name='contrast.api.dtm.Message.poll', index=21,
      number=27, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='noop', full_name='contrast.api.dtm.Message.noop', index=22,
      number=24, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='event', full_name='contrast.api.dtm.Message.event',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=32,
  serialized_end=957,
)


_POLL = _descriptor.Descriptor(
  name='Poll',
  full_name='contrast.api.dtm.Poll',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=959,
  serialized_end=965,
)


_NOOP = _descriptor.Descriptor(
  name='Noop',
  full_name='contrast.api.dtm.Noop',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='flag', full_name='contrast.api.dtm.Noop.flag', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\030\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=967,
  serialized_end=991,
)


_SERVERACTIVITY_LOGENHANCERSENTRY = _descriptor.Descriptor(
  name='LogEnhancersEntry',
  full_name='contrast.api.dtm.ServerActivity.LogEnhancersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.ServerActivity.LogEnhancersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.ServerActivity.LogEnhancersEntry.value', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1239,
  serialized_end=1290,
)

_SERVERACTIVITY_IPBLACKLISTSENTRY = _descriptor.Descriptor(
  name='IpBlacklistsEntry',
  full_name='contrast.api.dtm.ServerActivity.IpBlacklistsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.ServerActivity.IpBlacklistsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.ServerActivity.IpBlacklistsEntry.value', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1292,
  serialized_end=1343,
)

_SERVERACTIVITY_IPDENYLISTSENTRY = _descriptor.Descriptor(
  name='IpDenylistsEntry',
  full_name='contrast.api.dtm.ServerActivity.IpDenylistsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.ServerActivity.IpDenylistsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.ServerActivity.IpDenylistsEntry.value', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1345,
  serialized_end=1395,
)

_SERVERACTIVITY = _descriptor.Descriptor(
  name='ServerActivity',
  full_name='contrast.api.dtm.ServerActivity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_enhancers', full_name='contrast.api.dtm.ServerActivity.log_enhancers', index=0,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ip_blacklists', full_name='contrast.api.dtm.ServerActivity.ip_blacklists', index=1,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ip_denylists', full_name='contrast.api.dtm.ServerActivity.ip_denylists', index=2,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SERVERACTIVITY_LOGENHANCERSENTRY, _SERVERACTIVITY_IPBLACKLISTSENTRY, _SERVERACTIVITY_IPDENYLISTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=994,
  serialized_end=1395,
)


_AGENTSTARTUP = _descriptor.Descriptor(
  name='AgentStartup',
  full_name='contrast.api.dtm.AgentStartup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='contrast.api.dtm.AgentStartup.version', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='environment', full_name='contrast.api.dtm.AgentStartup.environment', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='contrast.api.dtm.AgentStartup.tags', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_name', full_name='contrast.api.dtm.AgentStartup.server_name', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_path', full_name='contrast.api.dtm.AgentStartup.server_path', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_type', full_name='contrast.api.dtm.AgentStartup.server_type', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_version', full_name='contrast.api.dtm.AgentStartup.server_version', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='heartbeat_ms', full_name='contrast.api.dtm.AgentStartup.heartbeat_ms', index=7,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_tags', full_name='contrast.api.dtm.AgentStartup.server_tags', index=8,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='application_tags', full_name='contrast.api.dtm.AgentStartup.application_tags', index=9,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='library_tags', full_name='contrast.api.dtm.AgentStartup.library_tags', index=10,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finding_tags', full_name='contrast.api.dtm.AgentStartup.finding_tags', index=11,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='agent_version', full_name='contrast.api.dtm.AgentStartup.agent_version', index=12,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1398,
  serialized_end=1695,
)


_APPLICATIONCREATE = _descriptor.Descriptor(
  name='ApplicationCreate',
  full_name='contrast.api.dtm.ApplicationCreate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='contrast.api.dtm.ApplicationCreate.group', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_version', full_name='contrast.api.dtm.ApplicationCreate.app_version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='contrast.api.dtm.ApplicationCreate.tags', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='contrast.api.dtm.ApplicationCreate.code', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='contrast.api.dtm.ApplicationCreate.metadata', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='contrast.api.dtm.ApplicationCreate.session_id', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_metadata', full_name='contrast.api.dtm.ApplicationCreate.session_metadata', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mode', full_name='contrast.api.dtm.ApplicationCreate.mode', index=7,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1698,
  serialized_end=1902,
)


_INSTRUMENTATIONMODE = _descriptor.Descriptor(
  name='InstrumentationMode',
  full_name='contrast.api.dtm.InstrumentationMode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='protect', full_name='contrast.api.dtm.InstrumentationMode.protect', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='assess', full_name='contrast.api.dtm.InstrumentationMode.assess', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1904,
  serialized_end=1958,
)


_APPLICATIONUPDATE_TECHNOLOGIESENTRY = _descriptor.Descriptor(
  name='TechnologiesEntry',
  full_name='contrast.api.dtm.ApplicationUpdate.TechnologiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.ApplicationUpdate.TechnologiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.ApplicationUpdate.TechnologiesEntry.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2346,
  serialized_end=2397,
)

_APPLICATIONUPDATE_LIBRARIESENTRY = _descriptor.Descriptor(
  name='LibrariesEntry',
  full_name='contrast.api.dtm.ApplicationUpdate.LibrariesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.ApplicationUpdate.LibrariesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.ApplicationUpdate.LibrariesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2399,
  serialized_end=2474,
)

_APPLICATIONUPDATE = _descriptor.Descriptor(
  name='ApplicationUpdate',
  full_name='contrast.api.dtm.ApplicationUpdate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='coverage_loc', full_name='contrast.api.dtm.ApplicationUpdate.coverage_loc', index=0,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='platform', full_name='contrast.api.dtm.ApplicationUpdate.platform', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='view', full_name='contrast.api.dtm.ApplicationUpdate.view', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='technologies', full_name='contrast.api.dtm.ApplicationUpdate.technologies', index=3,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='components', full_name='contrast.api.dtm.ApplicationUpdate.components', index=4,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='libraries', full_name='contrast.api.dtm.ApplicationUpdate.libraries', index=5,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='routes', full_name='contrast.api.dtm.ApplicationUpdate.routes', index=6,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_APPLICATIONUPDATE_TECHNOLOGIESENTRY, _APPLICATIONUPDATE_LIBRARIESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1961,
  serialized_end=2474,
)


_ACTIVITY_TECHNOLOGIESENTRY = _descriptor.Descriptor(
  name='TechnologiesEntry',
  full_name='contrast.api.dtm.Activity.TechnologiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.Activity.TechnologiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.Activity.TechnologiesEntry.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2346,
  serialized_end=2397,
)

_ACTIVITY_LIBRARYUSAGESENTRY = _descriptor.Descriptor(
  name='LibraryUsagesEntry',
  full_name='contrast.api.dtm.Activity.LibraryUsagesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.Activity.LibraryUsagesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.Activity.LibraryUsagesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3198,
  serialized_end=3288,
)

_ACTIVITY_DYNAMICSOURCESENTRY = _descriptor.Descriptor(
  name='DynamicSourcesEntry',
  full_name='contrast.api.dtm.Activity.DynamicSourcesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.Activity.DynamicSourcesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.Activity.DynamicSourcesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3290,
  serialized_end=3376,
)

_ACTIVITY = _descriptor.Descriptor(
  name='Activity',
  full_name='contrast.api.dtm.Activity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='http_request', full_name='contrast.api.dtm.Activity.http_request', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='http_response', full_name='contrast.api.dtm.Activity.http_response', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='duration_ms', full_name='contrast.api.dtm.Activity.duration_ms', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query_count', full_name='contrast.api.dtm.Activity.query_count', index=3,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='technologies', full_name='contrast.api.dtm.Activity.technologies', index=4,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='library_usages', full_name='contrast.api.dtm.Activity.library_usages', index=5,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='architectures', full_name='contrast.api.dtm.Activity.architectures', index=6,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='routes', full_name='contrast.api.dtm.Activity.routes', index=7,
      number=19, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='results', full_name='contrast.api.dtm.Activity.results', index=8,
      number=12, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='findings', full_name='contrast.api.dtm.Activity.findings', index=9,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='controls', full_name='contrast.api.dtm.Activity.controls', index=10,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finding_tags', full_name='contrast.api.dtm.Activity.finding_tags', index=11,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dynamic_sources', full_name='contrast.api.dtm.Activity.dynamic_sources', index=12,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ACTIVITY_TECHNOLOGIESENTRY, _ACTIVITY_LIBRARYUSAGESENTRY, _ACTIVITY_DYNAMICSOURCESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2477,
  serialized_end=3376,
)


_FINDING_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='contrast.api.dtm.Finding.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.Finding.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.Finding.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3698,
  serialized_end=3747,
)

_FINDING = _descriptor.Descriptor(
  name='Finding',
  full_name='contrast.api.dtm.Finding',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_code', full_name='contrast.api.dtm.Finding.hash_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='platform', full_name='contrast.api.dtm.Finding.platform', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rule_id', full_name='contrast.api.dtm.Finding.rule_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evidence', full_name='contrast.api.dtm.Finding.evidence', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='contrast.api.dtm.Finding.properties', index=4,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='events', full_name='contrast.api.dtm.Finding.events', index=5,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='preflight', full_name='contrast.api.dtm.Finding.preflight', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='contrast.api.dtm.Finding.tags', index=7,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='contrast.api.dtm.Finding.version', index=8,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='routes', full_name='contrast.api.dtm.Finding.routes', index=9,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='contrast.api.dtm.Finding.session_id', index=10,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_FINDING_PROPERTIESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3379,
  serialized_end=3747,
)


_DYNAMICSOURCE_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='contrast.api.dtm.DynamicSource.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.DynamicSource.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.DynamicSource.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3698,
  serialized_end=3747,
)

_DYNAMICSOURCE = _descriptor.Descriptor(
  name='DynamicSource',
  full_name='contrast.api.dtm.DynamicSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='class_name', full_name='contrast.api.dtm.DynamicSource.class_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method_name', full_name='contrast.api.dtm.DynamicSource.method_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='instance_method', full_name='contrast.api.dtm.DynamicSource.instance_method', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target', full_name='contrast.api.dtm.DynamicSource.target', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='events', full_name='contrast.api.dtm.DynamicSource.events', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='contrast.api.dtm.DynamicSource.properties', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DYNAMICSOURCE_PROPERTIESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3750,
  serialized_end=4013,
)


_POSSIBLESECURITYCONTROL = _descriptor.Descriptor(
  name='PossibleSecurityControl',
  full_name='contrast.api.dtm.PossibleSecurityControl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='api', full_name='contrast.api.dtm.PossibleSecurityControl.api', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='contrast.api.dtm.PossibleSecurityControl.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='traits', full_name='contrast.api.dtm.PossibleSecurityControl.traits', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4015,
  serialized_end=4083,
)


_VIEW = _descriptor.Descriptor(
  name='View',
  full_name='contrast.api.dtm.View',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='accessible', full_name='contrast.api.dtm.View.accessible', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='total_view_fields', full_name='contrast.api.dtm.View.total_view_fields', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_count', full_name='contrast.api.dtm.View.line_count', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4085,
  serialized_end=4158,
)


_PLATFORM = _descriptor.Descriptor(
  name='Platform',
  full_name='contrast.api.dtm.Platform',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='major', full_name='contrast.api.dtm.Platform.major', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='minor', full_name='contrast.api.dtm.Platform.minor', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='build', full_name='contrast.api.dtm.Platform.build', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_pack', full_name='contrast.api.dtm.Platform.service_pack', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updates', full_name='contrast.api.dtm.Platform.updates', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4160,
  serialized_end=4254,
)


_ARCHITECTURECOMPONENT = _descriptor.Descriptor(
  name='ArchitectureComponent',
  full_name='contrast.api.dtm.ArchitectureComponent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='contrast.api.dtm.ArchitectureComponent.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vendor', full_name='contrast.api.dtm.ArchitectureComponent.vendor', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='contrast.api.dtm.ArchitectureComponent.url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='remote_host', full_name='contrast.api.dtm.ArchitectureComponent.remote_host', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='remote_port', full_name='contrast.api.dtm.ArchitectureComponent.remote_port', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4256,
  serialized_end=4364,
)


_LIBRARY = _descriptor.Descriptor(
  name='Library',
  full_name='contrast.api.dtm.Library',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_code', full_name='contrast.api.dtm.Library.hash_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_path', full_name='contrast.api.dtm.Library.file_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='contrast.api.dtm.Library.url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='contrast.api.dtm.Library.version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tloc', full_name='contrast.api.dtm.Library.tloc', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='class_count', full_name='contrast.api.dtm.Library.class_count', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='used_class_count', full_name='contrast.api.dtm.Library.used_class_count', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='internal_ms', full_name='contrast.api.dtm.Library.internal_ms', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='external_ms', full_name='contrast.api.dtm.Library.external_ms', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='manifest', full_name='contrast.api.dtm.Library.manifest', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='contrast.api.dtm.Library.tags', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4367,
  serialized_end=4583,
)


_LIBRARYUSAGEUPDATE_CLASSNAMESENTRY = _descriptor.Descriptor(
  name='ClassNamesEntry',
  full_name='contrast.api.dtm.LibraryUsageUpdate.ClassNamesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.LibraryUsageUpdate.ClassNamesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.LibraryUsageUpdate.ClassNamesEntry.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4717,
  serialized_end=4766,
)

_LIBRARYUSAGEUPDATE = _descriptor.Descriptor(
  name='LibraryUsageUpdate',
  full_name='contrast.api.dtm.LibraryUsageUpdate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='contrast.api.dtm.LibraryUsageUpdate.count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hash_code', full_name='contrast.api.dtm.LibraryUsageUpdate.hash_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='class_names', full_name='contrast.api.dtm.LibraryUsageUpdate.class_names', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_LIBRARYUSAGEUPDATE_CLASSNAMESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4586,
  serialized_end=4766,
)


_ADDRESS = _descriptor.Descriptor(
  name='Address',
  full_name='contrast.api.dtm.Address',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip', full_name='contrast.api.dtm.Address.ip', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host', full_name='contrast.api.dtm.Address.host', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='contrast.api.dtm.Address.port', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4768,
  serialized_end=4817,
)


_RAWREQUEST = _descriptor.Descriptor(
  name='RawRequest',
  full_name='contrast.api.dtm.RawRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.RawRequest.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_ms', full_name='contrast.api.dtm.RawRequest.timestamp_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_line', full_name='contrast.api.dtm.RawRequest.request_line', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='normalized_uri', full_name='contrast.api.dtm.RawRequest.normalized_uri', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_ip', full_name='contrast.api.dtm.RawRequest.client_ip', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_ip_version', full_name='contrast.api.dtm.RawRequest.client_ip_version', index=5,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_port', full_name='contrast.api.dtm.RawRequest.client_port', index=6,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_ip', full_name='contrast.api.dtm.RawRequest.server_ip', index=7,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_ip_version', full_name='contrast.api.dtm.RawRequest.server_ip_version', index=8,
      number=11, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_port', full_name='contrast.api.dtm.RawRequest.server_port', index=9,
      number=12, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_headers', full_name='contrast.api.dtm.RawRequest.request_headers', index=10,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='multipart_headers', full_name='contrast.api.dtm.RawRequest.multipart_headers', index=11,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_body', full_name='contrast.api.dtm.RawRequest.request_body', index=12,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4820,
  serialized_end=5182,
)


_RAWRESPONSE = _descriptor.Descriptor(
  name='RawResponse',
  full_name='contrast.api.dtm.RawResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.RawResponse.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_ms', full_name='contrast.api.dtm.RawResponse.timestamp_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_code', full_name='contrast.api.dtm.RawResponse.response_code', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_headers', full_name='contrast.api.dtm.RawResponse.response_headers', index=3,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_body', full_name='contrast.api.dtm.RawResponse.response_body', index=4,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5185,
  serialized_end=5336,
)


_HTTPREQUEST_NORMALIZEDREQUESTPARAMSENTRY = _descriptor.Descriptor(
  name='NormalizedRequestParamsEntry',
  full_name='contrast.api.dtm.HttpRequest.NormalizedRequestParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.HttpRequest.NormalizedRequestParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.HttpRequest.NormalizedRequestParamsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6363,
  serialized_end=6449,
)

_HTTPREQUEST_REQUESTHEADERSENTRY = _descriptor.Descriptor(
  name='RequestHeadersEntry',
  full_name='contrast.api.dtm.HttpRequest.RequestHeadersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.HttpRequest.RequestHeadersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.HttpRequest.RequestHeadersEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6451,
  serialized_end=6504,
)

_HTTPREQUEST_NORMALIZEDREQUESTHEADERSENTRY = _descriptor.Descriptor(
  name='NormalizedRequestHeadersEntry',
  full_name='contrast.api.dtm.HttpRequest.NormalizedRequestHeadersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.HttpRequest.NormalizedRequestHeadersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.HttpRequest.NormalizedRequestHeadersEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6506,
  serialized_end=6593,
)

_HTTPREQUEST_NORMALIZEDCOOKIESENTRY = _descriptor.Descriptor(
  name='NormalizedCookiesEntry',
  full_name='contrast.api.dtm.HttpRequest.NormalizedCookiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.HttpRequest.NormalizedCookiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.HttpRequest.NormalizedCookiesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6595,
  serialized_end=6675,
)

_HTTPREQUEST = _descriptor.Descriptor(
  name='HttpRequest',
  full_name='contrast.api.dtm.HttpRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.HttpRequest.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_ms', full_name='contrast.api.dtm.HttpRequest.timestamp_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender', full_name='contrast.api.dtm.HttpRequest.sender', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='receiver', full_name='contrast.api.dtm.HttpRequest.receiver', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='protocol', full_name='contrast.api.dtm.HttpRequest.protocol', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='contrast.api.dtm.HttpRequest.version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method', full_name='contrast.api.dtm.HttpRequest.method', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw', full_name='contrast.api.dtm.HttpRequest.raw', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parsed_connection', full_name='contrast.api.dtm.HttpRequest.parsed_connection', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uri', full_name='contrast.api.dtm.HttpRequest.uri', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='normalized_uri', full_name='contrast.api.dtm.HttpRequest.normalized_uri', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='contrast.api.dtm.HttpRequest.context', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='contrast.api.dtm.HttpRequest.path', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query_string', full_name='contrast.api.dtm.HttpRequest.query_string', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='normalized_request_params', full_name='contrast.api.dtm.HttpRequest.normalized_request_params', index=14,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='multipart_headers', full_name='contrast.api.dtm.HttpRequest.multipart_headers', index=15,
      number=51, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_headers', full_name='contrast.api.dtm.HttpRequest.request_headers', index=16,
      number=20, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parsed_request_headers', full_name='contrast.api.dtm.HttpRequest.parsed_request_headers', index=17,
      number=21, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='normalized_request_headers', full_name='contrast.api.dtm.HttpRequest.normalized_request_headers', index=18,
      number=22, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='normalized_cookies', full_name='contrast.api.dtm.HttpRequest.normalized_cookies', index=19,
      number=23, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_body', full_name='contrast.api.dtm.HttpRequest.request_body', index=20,
      number=30, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parsed_request_body', full_name='contrast.api.dtm.HttpRequest.parsed_request_body', index=21,
      number=31, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='document_type', full_name='contrast.api.dtm.HttpRequest.document_type', index=22,
      number=32, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_body_binary', full_name='contrast.api.dtm.HttpRequest.request_body_binary', index=23,
      number=33, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='authentication', full_name='contrast.api.dtm.HttpRequest.authentication', index=24,
      number=40, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='principal', full_name='contrast.api.dtm.HttpRequest.principal', index=25,
      number=41, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_info', full_name='contrast.api.dtm.HttpRequest.server_info', index=26,
      number=50, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_HTTPREQUEST_NORMALIZEDREQUESTPARAMSENTRY, _HTTPREQUEST_REQUESTHEADERSENTRY, _HTTPREQUEST_NORMALIZEDREQUESTHEADERSENTRY, _HTTPREQUEST_NORMALIZEDCOOKIESENTRY, ],
  enum_types=[
    _HTTPREQUEST_AUTHENTICATION,
    _HTTPREQUEST_DOCUMENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5339,
  serialized_end=6834,
)


_RPCSTATUS = _descriptor.Descriptor(
  name='RpcStatus',
  full_name='contrast.api.dtm.RpcStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='contrast.api.dtm.RpcStatus.code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='msg', full_name='contrast.api.dtm.RpcStatus.msg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RPCSTATUS_CODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6836,
  serialized_end=6937,
)


_SQLISINKREQUEST = _descriptor.Descriptor(
  name='SqliSinkRequest',
  full_name='contrast.api.dtm.SqliSinkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.SqliSinkRequest.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dialect', full_name='contrast.api.dtm.SqliSinkRequest.dialect', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='contrast.api.dtm.SqliSinkRequest.query', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SQLISINKREQUEST_DIALECT,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6940,
  serialized_end=7097,
)


_SQLISINKRESPONSE = _descriptor.Descriptor(
  name='SqliSinkResponse',
  full_name='contrast.api.dtm.SqliSinkResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='contrast.api.dtm.SqliSinkResponse.status', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_input', full_name='contrast.api.dtm.SqliSinkResponse.user_input', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sqli', full_name='contrast.api.dtm.SqliSinkResponse.sqli', index=2,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sqli_semantic', full_name='contrast.api.dtm.SqliSinkResponse.sqli_semantic', index=3,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='finding', full_name='contrast.api.dtm.SqliSinkResponse.finding',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=7100,
  serialized_end=7358,
)


_HTTPRESPONSE_RESPONSEHEADERSENTRY = _descriptor.Descriptor(
  name='ResponseHeadersEntry',
  full_name='contrast.api.dtm.HttpResponse.ResponseHeadersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.HttpResponse.ResponseHeadersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.HttpResponse.ResponseHeadersEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7810,
  serialized_end=7864,
)

_HTTPRESPONSE_NORMALIZEDRESPONSEHEADERSENTRY = _descriptor.Descriptor(
  name='NormalizedResponseHeadersEntry',
  full_name='contrast.api.dtm.HttpResponse.NormalizedResponseHeadersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.HttpResponse.NormalizedResponseHeadersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.HttpResponse.NormalizedResponseHeadersEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7866,
  serialized_end=7954,
)

_HTTPRESPONSE = _descriptor.Descriptor(
  name='HttpResponse',
  full_name='contrast.api.dtm.HttpResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.HttpResponse.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_ms', full_name='contrast.api.dtm.HttpResponse.timestamp_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_code', full_name='contrast.api.dtm.HttpResponse.response_code', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_headers', full_name='contrast.api.dtm.HttpResponse.response_headers', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parsed_response_headers', full_name='contrast.api.dtm.HttpResponse.parsed_response_headers', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='normalized_response_headers', full_name='contrast.api.dtm.HttpResponse.normalized_response_headers', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_body', full_name='contrast.api.dtm.HttpResponse.response_body', index=6,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parsed_response_body', full_name='contrast.api.dtm.HttpResponse.parsed_response_body', index=7,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='document_type', full_name='contrast.api.dtm.HttpResponse.document_type', index=8,
      number=12, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_body_binary', full_name='contrast.api.dtm.HttpResponse.response_body_binary', index=9,
      number=13, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_HTTPRESPONSE_RESPONSEHEADERSENTRY, _HTTPRESPONSE_NORMALIZEDRESPONSEHEADERSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7361,
  serialized_end=7954,
)


_SIMPLEPAIR = _descriptor.Descriptor(
  name='SimplePair',
  full_name='contrast.api.dtm.SimplePair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.SimplePair.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.SimplePair.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7956,
  serialized_end=7996,
)


_PAIR = _descriptor.Descriptor(
  name='Pair',
  full_name='contrast.api.dtm.Pair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.Pair.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='contrast.api.dtm.Pair.values', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7998,
  serialized_end=8033,
)


_ATTACKRESULT = _descriptor.Descriptor(
  name='AttackResult',
  full_name='contrast.api.dtm.AttackResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='response', full_name='contrast.api.dtm.AttackResult.response', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rule_id', full_name='contrast.api.dtm.AttackResult.rule_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='samples', full_name='contrast.api.dtm.AttackResult.samples', index=2,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='contrast.api.dtm.AttackResult.tags', index=3,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='acc_updated_ms', full_name='contrast.api.dtm.AttackResult.acc_updated_ms', index=4,
      number=16, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='acc_limit', full_name='contrast.api.dtm.AttackResult.acc_limit', index=5,
      number=17, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='acc_window_ms', full_name='contrast.api.dtm.AttackResult.acc_window_ms', index=6,
      number=18, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='acc_count', full_name='contrast.api.dtm.AttackResult.acc_count', index=7,
      number=19, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='acc_mode', full_name='contrast.api.dtm.AttackResult.acc_mode', index=8,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='acc_msg', full_name='contrast.api.dtm.AttackResult.acc_msg', index=9,
      number=22, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ATTACKRESULT_RESPONSETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8036,
  serialized_end=8432,
)


_RASPRULESAMPLE = _descriptor.Descriptor(
  name='RaspRuleSample',
  full_name='contrast.api.dtm.RaspRuleSample',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp_ms', full_name='contrast.api.dtm.RaspRuleSample.timestamp_ms', index=0,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_trace_elements', full_name='contrast.api.dtm.RaspRuleSample.stack_trace_elements', index=1,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_input', full_name='contrast.api.dtm.RaspRuleSample.user_input', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='brute_force', full_name='contrast.api.dtm.RaspRuleSample.brute_force', index=3,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bot_blocker', full_name='contrast.api.dtm.RaspRuleSample.bot_blocker', index=4,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cmdi', full_name='contrast.api.dtm.RaspRuleSample.cmdi', index=5,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='csrf', full_name='contrast.api.dtm.RaspRuleSample.csrf', index=6,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cve', full_name='contrast.api.dtm.RaspRuleSample.cve', index=7,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='untrusted_deserialization', full_name='contrast.api.dtm.RaspRuleSample.untrusted_deserialization', index=8,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='el_injection', full_name='contrast.api.dtm.RaspRuleSample.el_injection', index=9,
      number=22, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ip_blacklist', full_name='contrast.api.dtm.RaspRuleSample.ip_blacklist', index=10,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ip_denylist', full_name='contrast.api.dtm.RaspRuleSample.ip_denylist', index=11,
      number=45, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mark_of_the_beast', full_name='contrast.api.dtm.RaspRuleSample.mark_of_the_beast', index=12,
      number=24, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='padding_oracle', full_name='contrast.api.dtm.RaspRuleSample.padding_oracle', index=13,
      number=25, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path_traversal', full_name='contrast.api.dtm.RaspRuleSample.path_traversal', index=14,
      number=26, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='re_dos', full_name='contrast.api.dtm.RaspRuleSample.re_dos', index=15,
      number=27, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sqli', full_name='contrast.api.dtm.RaspRuleSample.sqli', index=16,
      number=28, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ssrf', full_name='contrast.api.dtm.RaspRuleSample.ssrf', index=17,
      number=29, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='virtual_patch', full_name='contrast.api.dtm.RaspRuleSample.virtual_patch', index=18,
      number=30, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='xss', full_name='contrast.api.dtm.RaspRuleSample.xss', index=19,
      number=31, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='xxe', full_name='contrast.api.dtm.RaspRuleSample.xxe', index=20,
      number=32, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='no_sqli', full_name='contrast.api.dtm.RaspRuleSample.no_sqli', index=21,
      number=33, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method_tampering', full_name='contrast.api.dtm.RaspRuleSample.method_tampering', index=22,
      number=34, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cve_2017_0898', full_name='contrast.api.dtm.RaspRuleSample.cve_2017_0898', index=23,
      number=35, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cve_2017_14033', full_name='contrast.api.dtm.RaspRuleSample.cve_2017_14033', index=24,
      number=36, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cve_2017_14064', full_name='contrast.api.dtm.RaspRuleSample.cve_2017_14064', index=25,
      number=37, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cve_2017_17405', full_name='contrast.api.dtm.RaspRuleSample.cve_2017_17405', index=26,
      number=38, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mod_security', full_name='contrast.api.dtm.RaspRuleSample.mod_security', index=27,
      number=39, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='contrast.api.dtm.RaspRuleSample.endpoint', index=28,
      number=40, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sqli_semantic', full_name='contrast.api.dtm.RaspRuleSample.sqli_semantic', index=29,
      number=41, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cmdi_semantic', full_name='contrast.api.dtm.RaspRuleSample.cmdi_semantic', index=30,
      number=42, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path_traversal_semantic', full_name='contrast.api.dtm.RaspRuleSample.path_traversal_semantic', index=31,
      number=43, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ssjs', full_name='contrast.api.dtm.RaspRuleSample.ssjs', index=32,
      number=44, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='details', full_name='contrast.api.dtm.RaspRuleSample.details',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=8435,
  serialized_end=10490,
)


_ENDPOINTBANDETAILS = _descriptor.Descriptor(
  name='EndpointBanDetails',
  full_name='contrast.api.dtm.EndpointBanDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip', full_name='contrast.api.dtm.EndpointBanDetails.ip', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_uri', full_name='contrast.api.dtm.EndpointBanDetails.request_uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_code', full_name='contrast.api.dtm.EndpointBanDetails.response_code', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10492,
  serialized_end=10568,
)


_MODSECURITYDETAILS = _descriptor.Descriptor(
  name='ModSecurityDetails',
  full_name='contrast.api.dtm.ModSecurityDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='contrast.api.dtm.ModSecurityDetails.transaction_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='contrast.api.dtm.ModSecurityDetails.status', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='logs', full_name='contrast.api.dtm.ModSecurityDetails.logs', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10570,
  serialized_end=10678,
)


_MODSECURITYLOG_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='contrast.api.dtm.ModSecurityLog.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.ModSecurityLog.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.ModSecurityLog.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10778,
  serialized_end=10823,
)

_MODSECURITYLOG = _descriptor.Descriptor(
  name='ModSecurityLog',
  full_name='contrast.api.dtm.ModSecurityLog',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='contrast.api.dtm.ModSecurityLog.message', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='contrast.api.dtm.ModSecurityLog.params', index=1,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MODSECURITYLOG_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10681,
  serialized_end=10823,
)


_BRUTEFORCEDETAILS = _descriptor.Descriptor(
  name='BruteForceDetails',
  full_name='contrast.api.dtm.BruteForceDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='contrast.api.dtm.BruteForceDetails.count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10825,
  serialized_end=10859,
)


_BOTBLOCKERDETAILS = _descriptor.Descriptor(
  name='BotBlockerDetails',
  full_name='contrast.api.dtm.BotBlockerDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bot', full_name='contrast.api.dtm.BotBlockerDetails.bot', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_agent', full_name='contrast.api.dtm.BotBlockerDetails.user_agent', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10861,
  serialized_end=10913,
)


_CMDINJECTIONDETAILS = _descriptor.Descriptor(
  name='CmdInjectionDetails',
  full_name='contrast.api.dtm.CmdInjectionDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='command', full_name='contrast.api.dtm.CmdInjectionDetails.command', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_idx', full_name='contrast.api.dtm.CmdInjectionDetails.start_idx', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_idx', full_name='contrast.api.dtm.CmdInjectionDetails.end_idx', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10915,
  serialized_end=10989,
)


_CMDINJECTIONSEMANTICANALYSISDETAILS = _descriptor.Descriptor(
  name='CmdInjectionSemanticAnalysisDetails',
  full_name='contrast.api.dtm.CmdInjectionSemanticAnalysisDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='command', full_name='contrast.api.dtm.CmdInjectionSemanticAnalysisDetails.command', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='findings', full_name='contrast.api.dtm.CmdInjectionSemanticAnalysisDetails.findings', index=1,
      number=2, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CMDINJECTIONSEMANTICANALYSISDETAILS_FINDING,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10992,
  serialized_end=11186,
)


_CSRFDETAILS = _descriptor.Descriptor(
  name='CsrfDetails',
  full_name='contrast.api.dtm.CsrfDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='contrast.api.dtm.CsrfDetails.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='expected', full_name='contrast.api.dtm.CsrfDetails.expected', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='presented', full_name='contrast.api.dtm.CsrfDetails.presented', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11188,
  serialized_end=11252,
)


_CVEDETAILS = _descriptor.Descriptor(
  name='CveDetails',
  full_name='contrast.api.dtm.CveDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='contrast.api.dtm.CveDetails.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.CveDetails.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cve', full_name='contrast.api.dtm.CveDetails.cve', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='library', full_name='contrast.api.dtm.CveDetails.library', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11254,
  serialized_end=11325,
)


_CVE20170898DETAILS = _descriptor.Descriptor(
  name='Cve20170898Details',
  full_name='contrast.api.dtm.Cve20170898Details',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='format_string', full_name='contrast.api.dtm.Cve20170898Details.format_string', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='malicious_format_group', full_name='contrast.api.dtm.Cve20170898Details.malicious_format_group', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision_argument', full_name='contrast.api.dtm.Cve20170898Details.precision_argument', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value_argument', full_name='contrast.api.dtm.Cve20170898Details.value_argument', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='contrast.api.dtm.Cve20170898Details.arguments', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ruby_version', full_name='contrast.api.dtm.Cve20170898Details.ruby_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11328,
  serialized_end=11496,
)


_CVE201714033DETAILS = _descriptor.Descriptor(
  name='Cve201714033Details',
  full_name='contrast.api.dtm.Cve201714033Details',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='asn1', full_name='contrast.api.dtm.Cve201714033Details.asn1', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ruby_version', full_name='contrast.api.dtm.Cve201714033Details.ruby_version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11498,
  serialized_end=11555,
)


_CVE201714064DETAILS = _descriptor.Descriptor(
  name='Cve201714064Details',
  full_name='contrast.api.dtm.Cve201714064Details',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='space', full_name='contrast.api.dtm.Cve201714064Details.space', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ruby_version', full_name='contrast.api.dtm.Cve201714064Details.ruby_version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='json_version', full_name='contrast.api.dtm.Cve201714064Details.json_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11557,
  serialized_end=11637,
)


_CVE201717405DETAILS = _descriptor.Descriptor(
  name='Cve201717405Details',
  full_name='contrast.api.dtm.Cve201717405Details',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='contrast.api.dtm.Cve201717405Details.method', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='contrast.api.dtm.Cve201717405Details.file_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ruby_version', full_name='contrast.api.dtm.Cve201717405Details.ruby_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11639,
  serialized_end=11717,
)


_HTTPMETHODTAMPERINGDETAILS = _descriptor.Descriptor(
  name='HttpMethodTamperingDetails',
  full_name='contrast.api.dtm.HttpMethodTamperingDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='contrast.api.dtm.HttpMethodTamperingDetails.method', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response_code', full_name='contrast.api.dtm.HttpMethodTamperingDetails.response_code', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11719,
  serialized_end=11786,
)


_UNTRUSTEDDESERIALIZATIONDETAILS = _descriptor.Descriptor(
  name='UntrustedDeserializationDetails',
  full_name='contrast.api.dtm.UntrustedDeserializationDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='command', full_name='contrast.api.dtm.UntrustedDeserializationDetails.command', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='deserializer', full_name='contrast.api.dtm.UntrustedDeserializationDetails.deserializer', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11788,
  serialized_end=11860,
)


_ELINJECTIONDETAILS = _descriptor.Descriptor(
  name='ElInjectionDetails',
  full_name='contrast.api.dtm.ElInjectionDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='attr_value', full_name='contrast.api.dtm.ElInjectionDetails.attr_value', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='contrast.api.dtm.ElInjectionDetails.version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='library_filename', full_name='contrast.api.dtm.ElInjectionDetails.library_filename', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11862,
  serialized_end=11945,
)


_IPBLACKLISTDETAILS = _descriptor.Descriptor(
  name='IpBlacklistDetails',
  full_name='contrast.api.dtm.IpBlacklistDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip', full_name='contrast.api.dtm.IpBlacklistDetails.ip', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.IpBlacklistDetails.uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\030\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11947,
  serialized_end=11997,
)


_IPDENYLISTDETAILS = _descriptor.Descriptor(
  name='IpDenylistDetails',
  full_name='contrast.api.dtm.IpDenylistDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip', full_name='contrast.api.dtm.IpDenylistDetails.ip', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.IpDenylistDetails.uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11999,
  serialized_end=12044,
)


_MARKOFTHEBEASTDETAILS = _descriptor.Descriptor(
  name='MarkOfTheBeastDetails',
  full_name='contrast.api.dtm.MarkOfTheBeastDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='number', full_name='contrast.api.dtm.MarkOfTheBeastDetails.number', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vendor', full_name='contrast.api.dtm.MarkOfTheBeastDetails.vendor', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='contrast.api.dtm.MarkOfTheBeastDetails.version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12046,
  serialized_end=12118,
)


_PADDINGORACLEDETAILS = _descriptor.Descriptor(
  name='PaddingOracleDetails',
  full_name='contrast.api.dtm.PaddingOracleDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='contrast.api.dtm.PaddingOracleDetails.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period', full_name='contrast.api.dtm.PaddingOracleDetails.period', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hits', full_name='contrast.api.dtm.PaddingOracleDetails.hits', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12120,
  serialized_end=12189,
)


_PATHTRAVERSALDETAILS = _descriptor.Descriptor(
  name='PathTraversalDetails',
  full_name='contrast.api.dtm.PathTraversalDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='contrast.api.dtm.PathTraversalDetails.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12191,
  serialized_end=12227,
)


_PATHTRAVERSALSEMANTICANALYSISDETAILS = _descriptor.Descriptor(
  name='PathTraversalSemanticAnalysisDetails',
  full_name='contrast.api.dtm.PathTraversalSemanticAnalysisDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='contrast.api.dtm.PathTraversalSemanticAnalysisDetails.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='findings', full_name='contrast.api.dtm.PathTraversalSemanticAnalysisDetails.findings', index=1,
      number=2, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PATHTRAVERSALSEMANTICANALYSISDETAILS_FINDING,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12230,
  serialized_end=12441,
)


_REDOSDETAILS = _descriptor.Descriptor(
  name='ReDosDetails',
  full_name='contrast.api.dtm.ReDosDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='times_accessed', full_name='contrast.api.dtm.ReDosDetails.times_accessed', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='execution_time_ms', full_name='contrast.api.dtm.ReDosDetails.execution_time_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pattern', full_name='contrast.api.dtm.ReDosDetails.pattern', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offending_string', full_name='contrast.api.dtm.ReDosDetails.offending_string', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12443,
  serialized_end=12551,
)


_SSJSINJECTIONDETAILS = _descriptor.Descriptor(
  name='SsjsInjectionDetails',
  full_name='contrast.api.dtm.SsjsInjectionDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code_string', full_name='contrast.api.dtm.SsjsInjectionDetails.code_string', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_idx', full_name='contrast.api.dtm.SsjsInjectionDetails.start_idx', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_idx', full_name='contrast.api.dtm.SsjsInjectionDetails.end_idx', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='boundary_overrun_idx', full_name='contrast.api.dtm.SsjsInjectionDetails.boundary_overrun_idx', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12553,
  serialized_end=12662,
)


_SQLINJECTIONDETAILS = _descriptor.Descriptor(
  name='SqlInjectionDetails',
  full_name='contrast.api.dtm.SqlInjectionDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_idx', full_name='contrast.api.dtm.SqlInjectionDetails.start_idx', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_idx', full_name='contrast.api.dtm.SqlInjectionDetails.end_idx', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='boundary_overrun_idx', full_name='contrast.api.dtm.SqlInjectionDetails.boundary_overrun_idx', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='input_boundary_idx', full_name='contrast.api.dtm.SqlInjectionDetails.input_boundary_idx', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='contrast.api.dtm.SqlInjectionDetails.query', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12665,
  serialized_end=12795,
)


_SQLINJECTIONSEMANTICANALYSISDETAILS = _descriptor.Descriptor(
  name='SqlInjectionSemanticAnalysisDetails',
  full_name='contrast.api.dtm.SqlInjectionSemanticAnalysisDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='contrast.api.dtm.SqlInjectionSemanticAnalysisDetails.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finding', full_name='contrast.api.dtm.SqlInjectionSemanticAnalysisDetails.finding', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SQLINJECTIONSEMANTICANALYSISDETAILS_FINDING,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12798,
  serialized_end=13016,
)


_NOSQLINJECTIONDETAILS = _descriptor.Descriptor(
  name='NoSqlInjectionDetails',
  full_name='contrast.api.dtm.NoSqlInjectionDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_idx', full_name='contrast.api.dtm.NoSqlInjectionDetails.start_idx', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_idx', full_name='contrast.api.dtm.NoSqlInjectionDetails.end_idx', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='boundary_overrun_idx', full_name='contrast.api.dtm.NoSqlInjectionDetails.boundary_overrun_idx', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='input_boundary_idx', full_name='contrast.api.dtm.NoSqlInjectionDetails.input_boundary_idx', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='contrast.api.dtm.NoSqlInjectionDetails.query', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13019,
  serialized_end=13151,
)


_SSRFDETAILS = _descriptor.Descriptor(
  name='SsrfDetails',
  full_name='contrast.api.dtm.SsrfDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='contrast.api.dtm.SsrfDetails.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13153,
  serialized_end=13179,
)


_VIRTUALPATCHDETAILS = _descriptor.Descriptor(
  name='VirtualPatchDetails',
  full_name='contrast.api.dtm.VirtualPatchDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='contrast.api.dtm.VirtualPatchDetails.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13181,
  serialized_end=13216,
)


_XSSDETAILS = _descriptor.Descriptor(
  name='XssDetails',
  full_name='contrast.api.dtm.XssDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='input', full_name='contrast.api.dtm.XssDetails.input', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='matches', full_name='contrast.api.dtm.XssDetails.matches', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13218,
  serialized_end=13290,
)


_XSSMATCH = _descriptor.Descriptor(
  name='XssMatch',
  full_name='contrast.api.dtm.XssMatch',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='evidence_start_ms', full_name='contrast.api.dtm.XssMatch.evidence_start_ms', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evidence', full_name='contrast.api.dtm.XssMatch.evidence', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='contrast.api.dtm.XssMatch.offset', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13292,
  serialized_end=13363,
)


_XXEDETAILS = _descriptor.Descriptor(
  name='XxeDetails',
  full_name='contrast.api.dtm.XxeDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='xml', full_name='contrast.api.dtm.XxeDetails.xml', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='declared_entities', full_name='contrast.api.dtm.XxeDetails.declared_entities', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entities_resolved', full_name='contrast.api.dtm.XxeDetails.entities_resolved', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13366,
  serialized_end=13503,
)


_XXEMATCH = _descriptor.Descriptor(
  name='XxeMatch',
  full_name='contrast.api.dtm.XxeMatch',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_idx', full_name='contrast.api.dtm.XxeMatch.start_idx', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_idx', full_name='contrast.api.dtm.XxeMatch.end_idx', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13505,
  serialized_end=13551,
)


_XXEWRAPPER = _descriptor.Descriptor(
  name='XxeWrapper',
  full_name='contrast.api.dtm.XxeWrapper',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='system_id', full_name='contrast.api.dtm.XxeWrapper.system_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='public_id', full_name='contrast.api.dtm.XxeWrapper.public_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13553,
  serialized_end=13603,
)


_USERINPUT = _descriptor.Descriptor(
  name='UserInput',
  full_name='contrast.api.dtm.UserInput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='contrast.api.dtm.UserInput.path', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='contrast.api.dtm.UserInput.key', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.UserInput.value', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='input_type', full_name='contrast.api.dtm.UserInput.input_type', index=3,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='document_type', full_name='contrast.api.dtm.UserInput.document_type', index=4,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='matcher_ids', full_name='contrast.api.dtm.UserInput.matcher_ids', index=5,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _USERINPUT_INPUTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13606,
  serialized_end=14198,
)


_STACKTRACEELEMENT = _descriptor.Descriptor(
  name='StackTraceElement',
  full_name='contrast.api.dtm.StackTraceElement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='declaring_class', full_name='contrast.api.dtm.StackTraceElement.declaring_class', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method_name', full_name='contrast.api.dtm.StackTraceElement.method_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='contrast.api.dtm.StackTraceElement.file_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_number', full_name='contrast.api.dtm.StackTraceElement.line_number', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14200,
  serialized_end=14305,
)


_TRACEEVENT = _descriptor.Descriptor(
  name='TraceEvent',
  full_name='contrast.api.dtm.TraceEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action', full_name='contrast.api.dtm.TraceEvent.action', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='contrast.api.dtm.TraceEvent.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_ms', full_name='contrast.api.dtm.TraceEvent.timestamp_ms', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='thread', full_name='contrast.api.dtm.TraceEvent.thread', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signature', full_name='contrast.api.dtm.TraceEvent.signature', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='field_name', full_name='contrast.api.dtm.TraceEvent.field_name', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='contrast.api.dtm.TraceEvent.context', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='contrast.api.dtm.TraceEvent.code', index=7,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object', full_name='contrast.api.dtm.TraceEvent.object', index=8,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ret', full_name='contrast.api.dtm.TraceEvent.ret', index=9,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='args', full_name='contrast.api.dtm.TraceEvent.args', index=10,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack', full_name='contrast.api.dtm.TraceEvent.stack', index=11,
      number=19, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event_sources', full_name='contrast.api.dtm.TraceEvent.event_sources', index=12,
      number=20, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='contrast.api.dtm.TraceEvent.source', index=13,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target', full_name='contrast.api.dtm.TraceEvent.target', index=14,
      number=22, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='taint_ranges', full_name='contrast.api.dtm.TraceEvent.taint_ranges', index=15,
      number=23, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object_id', full_name='contrast.api.dtm.TraceEvent.object_id', index=16,
      number=24, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent_object_ids', full_name='contrast.api.dtm.TraceEvent.parent_object_ids', index=17,
      number=25, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TRACEEVENT_ACTION,
    _TRACEEVENT_EVENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14308,
  serialized_end=15236,
)


_TRACEEVENTSIGNATURE = _descriptor.Descriptor(
  name='TraceEventSignature',
  full_name='contrast.api.dtm.TraceEventSignature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_type', full_name='contrast.api.dtm.TraceEventSignature.return_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='class_name', full_name='contrast.api.dtm.TraceEventSignature.class_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method_name', full_name='contrast.api.dtm.TraceEventSignature.method_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arg_types', full_name='contrast.api.dtm.TraceEventSignature.arg_types', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='constructor', full_name='contrast.api.dtm.TraceEventSignature.constructor', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='void_method', full_name='contrast.api.dtm.TraceEventSignature.void_method', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='flags', full_name='contrast.api.dtm.TraceEventSignature.flags', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15239,
  serialized_end=15398,
)


_TRACEEVENTSOURCE = _descriptor.Descriptor(
  name='TraceEventSource',
  full_name='contrast.api.dtm.TraceEventSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='contrast.api.dtm.TraceEventSource.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='contrast.api.dtm.TraceEventSource.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15400,
  serialized_end=15446,
)


_TRACEEVENTOBJECT = _descriptor.Descriptor(
  name='TraceEventObject',
  full_name='contrast.api.dtm.TraceEventObject',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tracked', full_name='contrast.api.dtm.TraceEventObject.tracked', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='contrast.api.dtm.TraceEventObject.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ranges', full_name='contrast.api.dtm.TraceEventObject.ranges', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15448,
  serialized_end=15549,
)


_TRACETAINTRANGE = _descriptor.Descriptor(
  name='TraceTaintRange',
  full_name='contrast.api.dtm.TraceTaintRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tag', full_name='contrast.api.dtm.TraceTaintRange.tag', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='range', full_name='contrast.api.dtm.TraceTaintRange.range', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15551,
  serialized_end=15596,
)


_PARENTOBJECTID = _descriptor.Descriptor(
  name='ParentObjectId',
  full_name='contrast.api.dtm.ParentObjectId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='contrast.api.dtm.ParentObjectId.id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15598,
  serialized_end=15626,
)


_TRACESTACK = _descriptor.Descriptor(
  name='TraceStack',
  full_name='contrast.api.dtm.TraceStack',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signature', full_name='contrast.api.dtm.TraceStack.signature', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='declaring_class', full_name='contrast.api.dtm.TraceStack.declaring_class', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method_name', full_name='contrast.api.dtm.TraceStack.method_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='contrast.api.dtm.TraceStack.file_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_number', full_name='contrast.api.dtm.TraceStack.line_number', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='contrast.api.dtm.TraceStack.type', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='eval', full_name='contrast.api.dtm.TraceStack.eval', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15629,
  serialized_end=15774,
)


_ROUTECOVERAGE = _descriptor.Descriptor(
  name='RouteCoverage',
  full_name='contrast.api.dtm.RouteCoverage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='verb', full_name='contrast.api.dtm.RouteCoverage.verb', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='contrast.api.dtm.RouteCoverage.url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='route', full_name='contrast.api.dtm.RouteCoverage.route', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='contrast.api.dtm.RouteCoverage.count', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15776,
  serialized_end=15848,
)


_OBSERVEDROUTE = _descriptor.Descriptor(
  name='ObservedRoute',
  full_name='contrast.api.dtm.ObservedRoute',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signature', full_name='contrast.api.dtm.ObservedRoute.signature', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='contrast.api.dtm.ObservedRoute.url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='verb', full_name='contrast.api.dtm.ObservedRoute.verb', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='contrast.api.dtm.ObservedRoute.session_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sources', full_name='contrast.api.dtm.ObservedRoute.sources', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15851,
  serialized_end=15989,
)


_CMDISINKREQUEST = _descriptor.Descriptor(
  name='CmdiSinkRequest',
  full_name='contrast.api.dtm.CmdiSinkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15991,
  serialized_end=16008,
)


_CMDISINKRESPONSE = _descriptor.Descriptor(
  name='CmdiSinkResponse',
  full_name='contrast.api.dtm.CmdiSinkResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16010,
  serialized_end=16028,
)


_PTSINKREQUEST = _descriptor.Descriptor(
  name='PtSinkRequest',
  full_name='contrast.api.dtm.PtSinkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16030,
  serialized_end=16045,
)


_PTSINKRESPONSE = _descriptor.Descriptor(
  name='PtSinkResponse',
  full_name='contrast.api.dtm.PtSinkResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16047,
  serialized_end=16063,
)

_MESSAGE.fields_by_name['agent_startup'].message_type = _AGENTSTARTUP
_MESSAGE.fields_by_name['application_create'].message_type = _APPLICATIONCREATE
_MESSAGE.fields_by_name['server_activity'].message_type = _SERVERACTIVITY
_MESSAGE.fields_by_name['application_update'].message_type = _APPLICATIONUPDATE
_MESSAGE.fields_by_name['activity'].message_type = _ACTIVITY
_MESSAGE.fields_by_name['observed_route'].message_type = _OBSERVEDROUTE
_MESSAGE.fields_by_name['request'].message_type = _RAWREQUEST
_MESSAGE.fields_by_name['prefilter'].message_type = _HTTPREQUEST
_MESSAGE.fields_by_name['postfilter'].message_type = _HTTPRESPONSE
_MESSAGE.fields_by_name['response'].message_type = _RAWRESPONSE
_MESSAGE.fields_by_name['sqli_sink'].message_type = _SQLISINKREQUEST
_MESSAGE.fields_by_name['poll'].message_type = _POLL
_MESSAGE.fields_by_name['noop'].message_type = _NOOP
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['agent_startup'])
_MESSAGE.fields_by_name['agent_startup'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['application_create'])
_MESSAGE.fields_by_name['application_create'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['server_activity'])
_MESSAGE.fields_by_name['server_activity'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['application_update'])
_MESSAGE.fields_by_name['application_update'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['activity'])
_MESSAGE.fields_by_name['activity'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['observed_route'])
_MESSAGE.fields_by_name['observed_route'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['request'])
_MESSAGE.fields_by_name['request'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['prefilter'])
_MESSAGE.fields_by_name['prefilter'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['postfilter'])
_MESSAGE.fields_by_name['postfilter'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['response'])
_MESSAGE.fields_by_name['response'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['sqli_sink'])
_MESSAGE.fields_by_name['sqli_sink'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['poll'])
_MESSAGE.fields_by_name['poll'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_MESSAGE.oneofs_by_name['event'].fields.append(
  _MESSAGE.fields_by_name['noop'])
_MESSAGE.fields_by_name['noop'].containing_oneof = _MESSAGE.oneofs_by_name['event']
_SERVERACTIVITY_LOGENHANCERSENTRY.containing_type = _SERVERACTIVITY
_SERVERACTIVITY_IPBLACKLISTSENTRY.containing_type = _SERVERACTIVITY
_SERVERACTIVITY_IPDENYLISTSENTRY.containing_type = _SERVERACTIVITY
_SERVERACTIVITY.fields_by_name['log_enhancers'].message_type = _SERVERACTIVITY_LOGENHANCERSENTRY
_SERVERACTIVITY.fields_by_name['ip_blacklists'].message_type = _SERVERACTIVITY_IPBLACKLISTSENTRY
_SERVERACTIVITY.fields_by_name['ip_denylists'].message_type = _SERVERACTIVITY_IPDENYLISTSENTRY
_APPLICATIONCREATE.fields_by_name['mode'].message_type = _INSTRUMENTATIONMODE
_APPLICATIONUPDATE_TECHNOLOGIESENTRY.containing_type = _APPLICATIONUPDATE
_APPLICATIONUPDATE_LIBRARIESENTRY.fields_by_name['value'].message_type = _LIBRARY
_APPLICATIONUPDATE_LIBRARIESENTRY.containing_type = _APPLICATIONUPDATE
_APPLICATIONUPDATE.fields_by_name['platform'].message_type = _PLATFORM
_APPLICATIONUPDATE.fields_by_name['view'].message_type = _VIEW
_APPLICATIONUPDATE.fields_by_name['technologies'].message_type = _APPLICATIONUPDATE_TECHNOLOGIESENTRY
_APPLICATIONUPDATE.fields_by_name['components'].message_type = _ARCHITECTURECOMPONENT
_APPLICATIONUPDATE.fields_by_name['libraries'].message_type = _APPLICATIONUPDATE_LIBRARIESENTRY
_APPLICATIONUPDATE.fields_by_name['routes'].message_type = _ROUTECOVERAGE
_ACTIVITY_TECHNOLOGIESENTRY.containing_type = _ACTIVITY
_ACTIVITY_LIBRARYUSAGESENTRY.fields_by_name['value'].message_type = _LIBRARYUSAGEUPDATE
_ACTIVITY_LIBRARYUSAGESENTRY.containing_type = _ACTIVITY
_ACTIVITY_DYNAMICSOURCESENTRY.fields_by_name['value'].message_type = _DYNAMICSOURCE
_ACTIVITY_DYNAMICSOURCESENTRY.containing_type = _ACTIVITY
_ACTIVITY.fields_by_name['http_request'].message_type = _HTTPREQUEST
_ACTIVITY.fields_by_name['http_response'].message_type = _HTTPRESPONSE
_ACTIVITY.fields_by_name['technologies'].message_type = _ACTIVITY_TECHNOLOGIESENTRY
_ACTIVITY.fields_by_name['library_usages'].message_type = _ACTIVITY_LIBRARYUSAGESENTRY
_ACTIVITY.fields_by_name['architectures'].message_type = _ARCHITECTURECOMPONENT
_ACTIVITY.fields_by_name['routes'].message_type = _ROUTECOVERAGE
_ACTIVITY.fields_by_name['results'].message_type = _ATTACKRESULT
_ACTIVITY.fields_by_name['findings'].message_type = _FINDING
_ACTIVITY.fields_by_name['controls'].message_type = _POSSIBLESECURITYCONTROL
_ACTIVITY.fields_by_name['dynamic_sources'].message_type = _ACTIVITY_DYNAMICSOURCESENTRY
_FINDING_PROPERTIESENTRY.containing_type = _FINDING
_FINDING.fields_by_name['properties'].message_type = _FINDING_PROPERTIESENTRY
_FINDING.fields_by_name['events'].message_type = _TRACEEVENT
_FINDING.fields_by_name['routes'].message_type = _ROUTECOVERAGE
_DYNAMICSOURCE_PROPERTIESENTRY.containing_type = _DYNAMICSOURCE
_DYNAMICSOURCE.fields_by_name['events'].message_type = _TRACEEVENT
_DYNAMICSOURCE.fields_by_name['properties'].message_type = _DYNAMICSOURCE_PROPERTIESENTRY
_LIBRARYUSAGEUPDATE_CLASSNAMESENTRY.containing_type = _LIBRARYUSAGEUPDATE
_LIBRARYUSAGEUPDATE.fields_by_name['class_names'].message_type = _LIBRARYUSAGEUPDATE_CLASSNAMESENTRY
_RAWREQUEST.fields_by_name['request_headers'].message_type = _SIMPLEPAIR
_RAWREQUEST.fields_by_name['multipart_headers'].message_type = _SIMPLEPAIR
_RAWRESPONSE.fields_by_name['response_headers'].message_type = _SIMPLEPAIR
_HTTPREQUEST_NORMALIZEDREQUESTPARAMSENTRY.fields_by_name['value'].message_type = _PAIR
_HTTPREQUEST_NORMALIZEDREQUESTPARAMSENTRY.containing_type = _HTTPREQUEST
_HTTPREQUEST_REQUESTHEADERSENTRY.containing_type = _HTTPREQUEST
_HTTPREQUEST_NORMALIZEDREQUESTHEADERSENTRY.fields_by_name['value'].message_type = _PAIR
_HTTPREQUEST_NORMALIZEDREQUESTHEADERSENTRY.containing_type = _HTTPREQUEST
_HTTPREQUEST_NORMALIZEDCOOKIESENTRY.fields_by_name['value'].message_type = _PAIR
_HTTPREQUEST_NORMALIZEDCOOKIESENTRY.containing_type = _HTTPREQUEST
_HTTPREQUEST.fields_by_name['sender'].message_type = _ADDRESS
_HTTPREQUEST.fields_by_name['receiver'].message_type = _ADDRESS
_HTTPREQUEST.fields_by_name['normalized_request_params'].message_type = _HTTPREQUEST_NORMALIZEDREQUESTPARAMSENTRY
_HTTPREQUEST.fields_by_name['multipart_headers'].message_type = _SIMPLEPAIR
_HTTPREQUEST.fields_by_name['request_headers'].message_type = _HTTPREQUEST_REQUESTHEADERSENTRY
_HTTPREQUEST.fields_by_name['normalized_request_headers'].message_type = _HTTPREQUEST_NORMALIZEDREQUESTHEADERSENTRY
_HTTPREQUEST.fields_by_name['normalized_cookies'].message_type = _HTTPREQUEST_NORMALIZEDCOOKIESENTRY
_HTTPREQUEST.fields_by_name['document_type'].enum_type = _HTTPREQUEST_DOCUMENTTYPE
_HTTPREQUEST.fields_by_name['authentication'].enum_type = _HTTPREQUEST_AUTHENTICATION
_HTTPREQUEST_AUTHENTICATION.containing_type = _HTTPREQUEST
_HTTPREQUEST_DOCUMENTTYPE.containing_type = _HTTPREQUEST
_RPCSTATUS.fields_by_name['code'].enum_type = _RPCSTATUS_CODE
_RPCSTATUS_CODE.containing_type = _RPCSTATUS
_SQLISINKREQUEST.fields_by_name['dialect'].enum_type = _SQLISINKREQUEST_DIALECT
_SQLISINKREQUEST_DIALECT.containing_type = _SQLISINKREQUEST
_SQLISINKRESPONSE.fields_by_name['status'].message_type = _RPCSTATUS
_SQLISINKRESPONSE.fields_by_name['user_input'].message_type = _USERINPUT
_SQLISINKRESPONSE.fields_by_name['sqli'].message_type = _SQLINJECTIONDETAILS
_SQLISINKRESPONSE.fields_by_name['sqli_semantic'].message_type = _SQLINJECTIONSEMANTICANALYSISDETAILS
_SQLISINKRESPONSE.oneofs_by_name['finding'].fields.append(
  _SQLISINKRESPONSE.fields_by_name['sqli'])
_SQLISINKRESPONSE.fields_by_name['sqli'].containing_oneof = _SQLISINKRESPONSE.oneofs_by_name['finding']
_SQLISINKRESPONSE.oneofs_by_name['finding'].fields.append(
  _SQLISINKRESPONSE.fields_by_name['sqli_semantic'])
_SQLISINKRESPONSE.fields_by_name['sqli_semantic'].containing_oneof = _SQLISINKRESPONSE.oneofs_by_name['finding']
_HTTPRESPONSE_RESPONSEHEADERSENTRY.containing_type = _HTTPRESPONSE
_HTTPRESPONSE_NORMALIZEDRESPONSEHEADERSENTRY.fields_by_name['value'].message_type = _PAIR
_HTTPRESPONSE_NORMALIZEDRESPONSEHEADERSENTRY.containing_type = _HTTPRESPONSE
_HTTPRESPONSE.fields_by_name['response_headers'].message_type = _HTTPRESPONSE_RESPONSEHEADERSENTRY
_HTTPRESPONSE.fields_by_name['normalized_response_headers'].message_type = _HTTPRESPONSE_NORMALIZEDRESPONSEHEADERSENTRY
_HTTPRESPONSE.fields_by_name['document_type'].enum_type = _HTTPREQUEST_DOCUMENTTYPE
_ATTACKRESULT.fields_by_name['response'].enum_type = _ATTACKRESULT_RESPONSETYPE
_ATTACKRESULT.fields_by_name['samples'].message_type = _RASPRULESAMPLE
_ATTACKRESULT_RESPONSETYPE.containing_type = _ATTACKRESULT
_RASPRULESAMPLE.fields_by_name['stack_trace_elements'].message_type = _STACKTRACEELEMENT
_RASPRULESAMPLE.fields_by_name['user_input'].message_type = _USERINPUT
_RASPRULESAMPLE.fields_by_name['brute_force'].message_type = _BRUTEFORCEDETAILS
_RASPRULESAMPLE.fields_by_name['bot_blocker'].message_type = _BOTBLOCKERDETAILS
_RASPRULESAMPLE.fields_by_name['cmdi'].message_type = _CMDINJECTIONDETAILS
_RASPRULESAMPLE.fields_by_name['csrf'].message_type = _CSRFDETAILS
_RASPRULESAMPLE.fields_by_name['cve'].message_type = _CVEDETAILS
_RASPRULESAMPLE.fields_by_name['untrusted_deserialization'].message_type = _UNTRUSTEDDESERIALIZATIONDETAILS
_RASPRULESAMPLE.fields_by_name['el_injection'].message_type = _ELINJECTIONDETAILS
_RASPRULESAMPLE.fields_by_name['ip_blacklist'].message_type = _IPBLACKLISTDETAILS
_RASPRULESAMPLE.fields_by_name['ip_denylist'].message_type = _IPDENYLISTDETAILS
_RASPRULESAMPLE.fields_by_name['mark_of_the_beast'].message_type = _MARKOFTHEBEASTDETAILS
_RASPRULESAMPLE.fields_by_name['padding_oracle'].message_type = _PADDINGORACLEDETAILS
_RASPRULESAMPLE.fields_by_name['path_traversal'].message_type = _PATHTRAVERSALDETAILS
_RASPRULESAMPLE.fields_by_name['re_dos'].message_type = _REDOSDETAILS
_RASPRULESAMPLE.fields_by_name['sqli'].message_type = _SQLINJECTIONDETAILS
_RASPRULESAMPLE.fields_by_name['ssrf'].message_type = _SSRFDETAILS
_RASPRULESAMPLE.fields_by_name['virtual_patch'].message_type = _VIRTUALPATCHDETAILS
_RASPRULESAMPLE.fields_by_name['xss'].message_type = _XSSDETAILS
_RASPRULESAMPLE.fields_by_name['xxe'].message_type = _XXEDETAILS
_RASPRULESAMPLE.fields_by_name['no_sqli'].message_type = _NOSQLINJECTIONDETAILS
_RASPRULESAMPLE.fields_by_name['method_tampering'].message_type = _HTTPMETHODTAMPERINGDETAILS
_RASPRULESAMPLE.fields_by_name['cve_2017_0898'].message_type = _CVE20170898DETAILS
_RASPRULESAMPLE.fields_by_name['cve_2017_14033'].message_type = _CVE201714033DETAILS
_RASPRULESAMPLE.fields_by_name['cve_2017_14064'].message_type = _CVE201714064DETAILS
_RASPRULESAMPLE.fields_by_name['cve_2017_17405'].message_type = _CVE201717405DETAILS
_RASPRULESAMPLE.fields_by_name['mod_security'].message_type = _MODSECURITYDETAILS
_RASPRULESAMPLE.fields_by_name['endpoint'].message_type = _ENDPOINTBANDETAILS
_RASPRULESAMPLE.fields_by_name['sqli_semantic'].message_type = _SQLINJECTIONSEMANTICANALYSISDETAILS
_RASPRULESAMPLE.fields_by_name['cmdi_semantic'].message_type = _CMDINJECTIONSEMANTICANALYSISDETAILS
_RASPRULESAMPLE.fields_by_name['path_traversal_semantic'].message_type = _PATHTRAVERSALSEMANTICANALYSISDETAILS
_RASPRULESAMPLE.fields_by_name['ssjs'].message_type = _SSJSINJECTIONDETAILS
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['brute_force'])
_RASPRULESAMPLE.fields_by_name['brute_force'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['bot_blocker'])
_RASPRULESAMPLE.fields_by_name['bot_blocker'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['cmdi'])
_RASPRULESAMPLE.fields_by_name['cmdi'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['csrf'])
_RASPRULESAMPLE.fields_by_name['csrf'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['cve'])
_RASPRULESAMPLE.fields_by_name['cve'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['untrusted_deserialization'])
_RASPRULESAMPLE.fields_by_name['untrusted_deserialization'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['el_injection'])
_RASPRULESAMPLE.fields_by_name['el_injection'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['ip_blacklist'])
_RASPRULESAMPLE.fields_by_name['ip_blacklist'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['ip_denylist'])
_RASPRULESAMPLE.fields_by_name['ip_denylist'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['mark_of_the_beast'])
_RASPRULESAMPLE.fields_by_name['mark_of_the_beast'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['padding_oracle'])
_RASPRULESAMPLE.fields_by_name['padding_oracle'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['path_traversal'])
_RASPRULESAMPLE.fields_by_name['path_traversal'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['re_dos'])
_RASPRULESAMPLE.fields_by_name['re_dos'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['sqli'])
_RASPRULESAMPLE.fields_by_name['sqli'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['ssrf'])
_RASPRULESAMPLE.fields_by_name['ssrf'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['virtual_patch'])
_RASPRULESAMPLE.fields_by_name['virtual_patch'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['xss'])
_RASPRULESAMPLE.fields_by_name['xss'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['xxe'])
_RASPRULESAMPLE.fields_by_name['xxe'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['no_sqli'])
_RASPRULESAMPLE.fields_by_name['no_sqli'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['method_tampering'])
_RASPRULESAMPLE.fields_by_name['method_tampering'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['cve_2017_0898'])
_RASPRULESAMPLE.fields_by_name['cve_2017_0898'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['cve_2017_14033'])
_RASPRULESAMPLE.fields_by_name['cve_2017_14033'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['cve_2017_14064'])
_RASPRULESAMPLE.fields_by_name['cve_2017_14064'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['cve_2017_17405'])
_RASPRULESAMPLE.fields_by_name['cve_2017_17405'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['mod_security'])
_RASPRULESAMPLE.fields_by_name['mod_security'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['endpoint'])
_RASPRULESAMPLE.fields_by_name['endpoint'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['sqli_semantic'])
_RASPRULESAMPLE.fields_by_name['sqli_semantic'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['cmdi_semantic'])
_RASPRULESAMPLE.fields_by_name['cmdi_semantic'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['path_traversal_semantic'])
_RASPRULESAMPLE.fields_by_name['path_traversal_semantic'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_RASPRULESAMPLE.oneofs_by_name['details'].fields.append(
  _RASPRULESAMPLE.fields_by_name['ssjs'])
_RASPRULESAMPLE.fields_by_name['ssjs'].containing_oneof = _RASPRULESAMPLE.oneofs_by_name['details']
_MODSECURITYDETAILS.fields_by_name['logs'].message_type = _MODSECURITYLOG
_MODSECURITYLOG_PARAMSENTRY.containing_type = _MODSECURITYLOG
_MODSECURITYLOG.fields_by_name['params'].message_type = _MODSECURITYLOG_PARAMSENTRY
_CMDINJECTIONSEMANTICANALYSISDETAILS.fields_by_name['findings'].enum_type = _CMDINJECTIONSEMANTICANALYSISDETAILS_FINDING
_CMDINJECTIONSEMANTICANALYSISDETAILS_FINDING.containing_type = _CMDINJECTIONSEMANTICANALYSISDETAILS
_PATHTRAVERSALSEMANTICANALYSISDETAILS.fields_by_name['findings'].enum_type = _PATHTRAVERSALSEMANTICANALYSISDETAILS_FINDING
_PATHTRAVERSALSEMANTICANALYSISDETAILS_FINDING.containing_type = _PATHTRAVERSALSEMANTICANALYSISDETAILS
_SQLINJECTIONSEMANTICANALYSISDETAILS.fields_by_name['finding'].enum_type = _SQLINJECTIONSEMANTICANALYSISDETAILS_FINDING
_SQLINJECTIONSEMANTICANALYSISDETAILS_FINDING.containing_type = _SQLINJECTIONSEMANTICANALYSISDETAILS
_XSSDETAILS.fields_by_name['matches'].message_type = _XSSMATCH
_XXEDETAILS.fields_by_name['declared_entities'].message_type = _XXEMATCH
_XXEDETAILS.fields_by_name['entities_resolved'].message_type = _XXEWRAPPER
_USERINPUT.fields_by_name['input_type'].enum_type = _USERINPUT_INPUTTYPE
_USERINPUT.fields_by_name['document_type'].enum_type = _HTTPREQUEST_DOCUMENTTYPE
_USERINPUT_INPUTTYPE.containing_type = _USERINPUT
_TRACEEVENT.fields_by_name['action'].enum_type = _TRACEEVENT_ACTION
_TRACEEVENT.fields_by_name['type'].enum_type = _TRACEEVENT_EVENTTYPE
_TRACEEVENT.fields_by_name['signature'].message_type = _TRACEEVENTSIGNATURE
_TRACEEVENT.fields_by_name['object'].message_type = _TRACEEVENTOBJECT
_TRACEEVENT.fields_by_name['ret'].message_type = _TRACEEVENTOBJECT
_TRACEEVENT.fields_by_name['args'].message_type = _TRACEEVENTOBJECT
_TRACEEVENT.fields_by_name['stack'].message_type = _TRACESTACK
_TRACEEVENT.fields_by_name['event_sources'].message_type = _TRACEEVENTSOURCE
_TRACEEVENT.fields_by_name['taint_ranges'].message_type = _TRACETAINTRANGE
_TRACEEVENT.fields_by_name['parent_object_ids'].message_type = _PARENTOBJECTID
_TRACEEVENT_ACTION.containing_type = _TRACEEVENT
_TRACEEVENT_EVENTTYPE.containing_type = _TRACEEVENT
_TRACEEVENTOBJECT.fields_by_name['ranges'].message_type = _TRACETAINTRANGE
_OBSERVEDROUTE.fields_by_name['sources'].message_type = _TRACEEVENTSOURCE
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['Poll'] = _POLL
DESCRIPTOR.message_types_by_name['Noop'] = _NOOP
DESCRIPTOR.message_types_by_name['ServerActivity'] = _SERVERACTIVITY
DESCRIPTOR.message_types_by_name['AgentStartup'] = _AGENTSTARTUP
DESCRIPTOR.message_types_by_name['ApplicationCreate'] = _APPLICATIONCREATE
DESCRIPTOR.message_types_by_name['InstrumentationMode'] = _INSTRUMENTATIONMODE
DESCRIPTOR.message_types_by_name['ApplicationUpdate'] = _APPLICATIONUPDATE
DESCRIPTOR.message_types_by_name['Activity'] = _ACTIVITY
DESCRIPTOR.message_types_by_name['Finding'] = _FINDING
DESCRIPTOR.message_types_by_name['DynamicSource'] = _DYNAMICSOURCE
DESCRIPTOR.message_types_by_name['PossibleSecurityControl'] = _POSSIBLESECURITYCONTROL
DESCRIPTOR.message_types_by_name['View'] = _VIEW
DESCRIPTOR.message_types_by_name['Platform'] = _PLATFORM
DESCRIPTOR.message_types_by_name['ArchitectureComponent'] = _ARCHITECTURECOMPONENT
DESCRIPTOR.message_types_by_name['Library'] = _LIBRARY
DESCRIPTOR.message_types_by_name['LibraryUsageUpdate'] = _LIBRARYUSAGEUPDATE
DESCRIPTOR.message_types_by_name['Address'] = _ADDRESS
DESCRIPTOR.message_types_by_name['RawRequest'] = _RAWREQUEST
DESCRIPTOR.message_types_by_name['RawResponse'] = _RAWRESPONSE
DESCRIPTOR.message_types_by_name['HttpRequest'] = _HTTPREQUEST
DESCRIPTOR.message_types_by_name['RpcStatus'] = _RPCSTATUS
DESCRIPTOR.message_types_by_name['SqliSinkRequest'] = _SQLISINKREQUEST
DESCRIPTOR.message_types_by_name['SqliSinkResponse'] = _SQLISINKRESPONSE
DESCRIPTOR.message_types_by_name['HttpResponse'] = _HTTPRESPONSE
DESCRIPTOR.message_types_by_name['SimplePair'] = _SIMPLEPAIR
DESCRIPTOR.message_types_by_name['Pair'] = _PAIR
DESCRIPTOR.message_types_by_name['AttackResult'] = _ATTACKRESULT
DESCRIPTOR.message_types_by_name['RaspRuleSample'] = _RASPRULESAMPLE
DESCRIPTOR.message_types_by_name['EndpointBanDetails'] = _ENDPOINTBANDETAILS
DESCRIPTOR.message_types_by_name['ModSecurityDetails'] = _MODSECURITYDETAILS
DESCRIPTOR.message_types_by_name['ModSecurityLog'] = _MODSECURITYLOG
DESCRIPTOR.message_types_by_name['BruteForceDetails'] = _BRUTEFORCEDETAILS
DESCRIPTOR.message_types_by_name['BotBlockerDetails'] = _BOTBLOCKERDETAILS
DESCRIPTOR.message_types_by_name['CmdInjectionDetails'] = _CMDINJECTIONDETAILS
DESCRIPTOR.message_types_by_name['CmdInjectionSemanticAnalysisDetails'] = _CMDINJECTIONSEMANTICANALYSISDETAILS
DESCRIPTOR.message_types_by_name['CsrfDetails'] = _CSRFDETAILS
DESCRIPTOR.message_types_by_name['CveDetails'] = _CVEDETAILS
DESCRIPTOR.message_types_by_name['Cve20170898Details'] = _CVE20170898DETAILS
DESCRIPTOR.message_types_by_name['Cve201714033Details'] = _CVE201714033DETAILS
DESCRIPTOR.message_types_by_name['Cve201714064Details'] = _CVE201714064DETAILS
DESCRIPTOR.message_types_by_name['Cve201717405Details'] = _CVE201717405DETAILS
DESCRIPTOR.message_types_by_name['HttpMethodTamperingDetails'] = _HTTPMETHODTAMPERINGDETAILS
DESCRIPTOR.message_types_by_name['UntrustedDeserializationDetails'] = _UNTRUSTEDDESERIALIZATIONDETAILS
DESCRIPTOR.message_types_by_name['ElInjectionDetails'] = _ELINJECTIONDETAILS
DESCRIPTOR.message_types_by_name['IpBlacklistDetails'] = _IPBLACKLISTDETAILS
DESCRIPTOR.message_types_by_name['IpDenylistDetails'] = _IPDENYLISTDETAILS
DESCRIPTOR.message_types_by_name['MarkOfTheBeastDetails'] = _MARKOFTHEBEASTDETAILS
DESCRIPTOR.message_types_by_name['PaddingOracleDetails'] = _PADDINGORACLEDETAILS
DESCRIPTOR.message_types_by_name['PathTraversalDetails'] = _PATHTRAVERSALDETAILS
DESCRIPTOR.message_types_by_name['PathTraversalSemanticAnalysisDetails'] = _PATHTRAVERSALSEMANTICANALYSISDETAILS
DESCRIPTOR.message_types_by_name['ReDosDetails'] = _REDOSDETAILS
DESCRIPTOR.message_types_by_name['SsjsInjectionDetails'] = _SSJSINJECTIONDETAILS
DESCRIPTOR.message_types_by_name['SqlInjectionDetails'] = _SQLINJECTIONDETAILS
DESCRIPTOR.message_types_by_name['SqlInjectionSemanticAnalysisDetails'] = _SQLINJECTIONSEMANTICANALYSISDETAILS
DESCRIPTOR.message_types_by_name['NoSqlInjectionDetails'] = _NOSQLINJECTIONDETAILS
DESCRIPTOR.message_types_by_name['SsrfDetails'] = _SSRFDETAILS
DESCRIPTOR.message_types_by_name['VirtualPatchDetails'] = _VIRTUALPATCHDETAILS
DESCRIPTOR.message_types_by_name['XssDetails'] = _XSSDETAILS
DESCRIPTOR.message_types_by_name['XssMatch'] = _XSSMATCH
DESCRIPTOR.message_types_by_name['XxeDetails'] = _XXEDETAILS
DESCRIPTOR.message_types_by_name['XxeMatch'] = _XXEMATCH
DESCRIPTOR.message_types_by_name['XxeWrapper'] = _XXEWRAPPER
DESCRIPTOR.message_types_by_name['UserInput'] = _USERINPUT
DESCRIPTOR.message_types_by_name['StackTraceElement'] = _STACKTRACEELEMENT
DESCRIPTOR.message_types_by_name['TraceEvent'] = _TRACEEVENT
DESCRIPTOR.message_types_by_name['TraceEventSignature'] = _TRACEEVENTSIGNATURE
DESCRIPTOR.message_types_by_name['TraceEventSource'] = _TRACEEVENTSOURCE
DESCRIPTOR.message_types_by_name['TraceEventObject'] = _TRACEEVENTOBJECT
DESCRIPTOR.message_types_by_name['TraceTaintRange'] = _TRACETAINTRANGE
DESCRIPTOR.message_types_by_name['ParentObjectId'] = _PARENTOBJECTID
DESCRIPTOR.message_types_by_name['TraceStack'] = _TRACESTACK
DESCRIPTOR.message_types_by_name['RouteCoverage'] = _ROUTECOVERAGE
DESCRIPTOR.message_types_by_name['ObservedRoute'] = _OBSERVEDROUTE
DESCRIPTOR.message_types_by_name['CmdiSinkRequest'] = _CMDISINKREQUEST
DESCRIPTOR.message_types_by_name['CmdiSinkResponse'] = _CMDISINKRESPONSE
DESCRIPTOR.message_types_by_name['PtSinkRequest'] = _PTSINKREQUEST
DESCRIPTOR.message_types_by_name['PtSinkResponse'] = _PTSINKRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Message)
  })
_sym_db.RegisterMessage(Message)

Poll = _reflection.GeneratedProtocolMessageType('Poll', (_message.Message,), {
  'DESCRIPTOR' : _POLL,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Poll)
  })
_sym_db.RegisterMessage(Poll)

Noop = _reflection.GeneratedProtocolMessageType('Noop', (_message.Message,), {
  'DESCRIPTOR' : _NOOP,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Noop)
  })
_sym_db.RegisterMessage(Noop)

ServerActivity = _reflection.GeneratedProtocolMessageType('ServerActivity', (_message.Message,), {

  'LogEnhancersEntry' : _reflection.GeneratedProtocolMessageType('LogEnhancersEntry', (_message.Message,), {
    'DESCRIPTOR' : _SERVERACTIVITY_LOGENHANCERSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.ServerActivity.LogEnhancersEntry)
    })
  ,

  'IpBlacklistsEntry' : _reflection.GeneratedProtocolMessageType('IpBlacklistsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SERVERACTIVITY_IPBLACKLISTSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.ServerActivity.IpBlacklistsEntry)
    })
  ,

  'IpDenylistsEntry' : _reflection.GeneratedProtocolMessageType('IpDenylistsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SERVERACTIVITY_IPDENYLISTSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.ServerActivity.IpDenylistsEntry)
    })
  ,
  'DESCRIPTOR' : _SERVERACTIVITY,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ServerActivity)
  })
_sym_db.RegisterMessage(ServerActivity)
_sym_db.RegisterMessage(ServerActivity.LogEnhancersEntry)
_sym_db.RegisterMessage(ServerActivity.IpBlacklistsEntry)
_sym_db.RegisterMessage(ServerActivity.IpDenylistsEntry)

AgentStartup = _reflection.GeneratedProtocolMessageType('AgentStartup', (_message.Message,), {
  'DESCRIPTOR' : _AGENTSTARTUP,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.AgentStartup)
  })
_sym_db.RegisterMessage(AgentStartup)

ApplicationCreate = _reflection.GeneratedProtocolMessageType('ApplicationCreate', (_message.Message,), {
  'DESCRIPTOR' : _APPLICATIONCREATE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ApplicationCreate)
  })
_sym_db.RegisterMessage(ApplicationCreate)

InstrumentationMode = _reflection.GeneratedProtocolMessageType('InstrumentationMode', (_message.Message,), {
  'DESCRIPTOR' : _INSTRUMENTATIONMODE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.InstrumentationMode)
  })
_sym_db.RegisterMessage(InstrumentationMode)

ApplicationUpdate = _reflection.GeneratedProtocolMessageType('ApplicationUpdate', (_message.Message,), {

  'TechnologiesEntry' : _reflection.GeneratedProtocolMessageType('TechnologiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _APPLICATIONUPDATE_TECHNOLOGIESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.ApplicationUpdate.TechnologiesEntry)
    })
  ,

  'LibrariesEntry' : _reflection.GeneratedProtocolMessageType('LibrariesEntry', (_message.Message,), {
    'DESCRIPTOR' : _APPLICATIONUPDATE_LIBRARIESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.ApplicationUpdate.LibrariesEntry)
    })
  ,
  'DESCRIPTOR' : _APPLICATIONUPDATE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ApplicationUpdate)
  })
_sym_db.RegisterMessage(ApplicationUpdate)
_sym_db.RegisterMessage(ApplicationUpdate.TechnologiesEntry)
_sym_db.RegisterMessage(ApplicationUpdate.LibrariesEntry)

Activity = _reflection.GeneratedProtocolMessageType('Activity', (_message.Message,), {

  'TechnologiesEntry' : _reflection.GeneratedProtocolMessageType('TechnologiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _ACTIVITY_TECHNOLOGIESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.Activity.TechnologiesEntry)
    })
  ,

  'LibraryUsagesEntry' : _reflection.GeneratedProtocolMessageType('LibraryUsagesEntry', (_message.Message,), {
    'DESCRIPTOR' : _ACTIVITY_LIBRARYUSAGESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.Activity.LibraryUsagesEntry)
    })
  ,

  'DynamicSourcesEntry' : _reflection.GeneratedProtocolMessageType('DynamicSourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _ACTIVITY_DYNAMICSOURCESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.Activity.DynamicSourcesEntry)
    })
  ,
  'DESCRIPTOR' : _ACTIVITY,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Activity)
  })
_sym_db.RegisterMessage(Activity)
_sym_db.RegisterMessage(Activity.TechnologiesEntry)
_sym_db.RegisterMessage(Activity.LibraryUsagesEntry)
_sym_db.RegisterMessage(Activity.DynamicSourcesEntry)

Finding = _reflection.GeneratedProtocolMessageType('Finding', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _FINDING_PROPERTIESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.Finding.PropertiesEntry)
    })
  ,
  'DESCRIPTOR' : _FINDING,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Finding)
  })
_sym_db.RegisterMessage(Finding)
_sym_db.RegisterMessage(Finding.PropertiesEntry)

DynamicSource = _reflection.GeneratedProtocolMessageType('DynamicSource', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _DYNAMICSOURCE_PROPERTIESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.DynamicSource.PropertiesEntry)
    })
  ,
  'DESCRIPTOR' : _DYNAMICSOURCE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.DynamicSource)
  })
_sym_db.RegisterMessage(DynamicSource)
_sym_db.RegisterMessage(DynamicSource.PropertiesEntry)

PossibleSecurityControl = _reflection.GeneratedProtocolMessageType('PossibleSecurityControl', (_message.Message,), {
  'DESCRIPTOR' : _POSSIBLESECURITYCONTROL,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.PossibleSecurityControl)
  })
_sym_db.RegisterMessage(PossibleSecurityControl)

View = _reflection.GeneratedProtocolMessageType('View', (_message.Message,), {
  'DESCRIPTOR' : _VIEW,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.View)
  })
_sym_db.RegisterMessage(View)

Platform = _reflection.GeneratedProtocolMessageType('Platform', (_message.Message,), {
  'DESCRIPTOR' : _PLATFORM,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Platform)
  })
_sym_db.RegisterMessage(Platform)

ArchitectureComponent = _reflection.GeneratedProtocolMessageType('ArchitectureComponent', (_message.Message,), {
  'DESCRIPTOR' : _ARCHITECTURECOMPONENT,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ArchitectureComponent)
  })
_sym_db.RegisterMessage(ArchitectureComponent)

Library = _reflection.GeneratedProtocolMessageType('Library', (_message.Message,), {
  'DESCRIPTOR' : _LIBRARY,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Library)
  })
_sym_db.RegisterMessage(Library)

LibraryUsageUpdate = _reflection.GeneratedProtocolMessageType('LibraryUsageUpdate', (_message.Message,), {

  'ClassNamesEntry' : _reflection.GeneratedProtocolMessageType('ClassNamesEntry', (_message.Message,), {
    'DESCRIPTOR' : _LIBRARYUSAGEUPDATE_CLASSNAMESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.LibraryUsageUpdate.ClassNamesEntry)
    })
  ,
  'DESCRIPTOR' : _LIBRARYUSAGEUPDATE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.LibraryUsageUpdate)
  })
_sym_db.RegisterMessage(LibraryUsageUpdate)
_sym_db.RegisterMessage(LibraryUsageUpdate.ClassNamesEntry)

Address = _reflection.GeneratedProtocolMessageType('Address', (_message.Message,), {
  'DESCRIPTOR' : _ADDRESS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Address)
  })
_sym_db.RegisterMessage(Address)

RawRequest = _reflection.GeneratedProtocolMessageType('RawRequest', (_message.Message,), {
  'DESCRIPTOR' : _RAWREQUEST,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.RawRequest)
  })
_sym_db.RegisterMessage(RawRequest)

RawResponse = _reflection.GeneratedProtocolMessageType('RawResponse', (_message.Message,), {
  'DESCRIPTOR' : _RAWRESPONSE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.RawResponse)
  })
_sym_db.RegisterMessage(RawResponse)

HttpRequest = _reflection.GeneratedProtocolMessageType('HttpRequest', (_message.Message,), {

  'NormalizedRequestParamsEntry' : _reflection.GeneratedProtocolMessageType('NormalizedRequestParamsEntry', (_message.Message,), {
    'DESCRIPTOR' : _HTTPREQUEST_NORMALIZEDREQUESTPARAMSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpRequest.NormalizedRequestParamsEntry)
    })
  ,

  'RequestHeadersEntry' : _reflection.GeneratedProtocolMessageType('RequestHeadersEntry', (_message.Message,), {
    'DESCRIPTOR' : _HTTPREQUEST_REQUESTHEADERSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpRequest.RequestHeadersEntry)
    })
  ,

  'NormalizedRequestHeadersEntry' : _reflection.GeneratedProtocolMessageType('NormalizedRequestHeadersEntry', (_message.Message,), {
    'DESCRIPTOR' : _HTTPREQUEST_NORMALIZEDREQUESTHEADERSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpRequest.NormalizedRequestHeadersEntry)
    })
  ,

  'NormalizedCookiesEntry' : _reflection.GeneratedProtocolMessageType('NormalizedCookiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _HTTPREQUEST_NORMALIZEDCOOKIESENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpRequest.NormalizedCookiesEntry)
    })
  ,
  'DESCRIPTOR' : _HTTPREQUEST,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpRequest)
  })
_sym_db.RegisterMessage(HttpRequest)
_sym_db.RegisterMessage(HttpRequest.NormalizedRequestParamsEntry)
_sym_db.RegisterMessage(HttpRequest.RequestHeadersEntry)
_sym_db.RegisterMessage(HttpRequest.NormalizedRequestHeadersEntry)
_sym_db.RegisterMessage(HttpRequest.NormalizedCookiesEntry)

RpcStatus = _reflection.GeneratedProtocolMessageType('RpcStatus', (_message.Message,), {
  'DESCRIPTOR' : _RPCSTATUS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.RpcStatus)
  })
_sym_db.RegisterMessage(RpcStatus)

SqliSinkRequest = _reflection.GeneratedProtocolMessageType('SqliSinkRequest', (_message.Message,), {
  'DESCRIPTOR' : _SQLISINKREQUEST,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.SqliSinkRequest)
  })
_sym_db.RegisterMessage(SqliSinkRequest)

SqliSinkResponse = _reflection.GeneratedProtocolMessageType('SqliSinkResponse', (_message.Message,), {
  'DESCRIPTOR' : _SQLISINKRESPONSE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.SqliSinkResponse)
  })
_sym_db.RegisterMessage(SqliSinkResponse)

HttpResponse = _reflection.GeneratedProtocolMessageType('HttpResponse', (_message.Message,), {

  'ResponseHeadersEntry' : _reflection.GeneratedProtocolMessageType('ResponseHeadersEntry', (_message.Message,), {
    'DESCRIPTOR' : _HTTPRESPONSE_RESPONSEHEADERSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpResponse.ResponseHeadersEntry)
    })
  ,

  'NormalizedResponseHeadersEntry' : _reflection.GeneratedProtocolMessageType('NormalizedResponseHeadersEntry', (_message.Message,), {
    'DESCRIPTOR' : _HTTPRESPONSE_NORMALIZEDRESPONSEHEADERSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpResponse.NormalizedResponseHeadersEntry)
    })
  ,
  'DESCRIPTOR' : _HTTPRESPONSE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpResponse)
  })
_sym_db.RegisterMessage(HttpResponse)
_sym_db.RegisterMessage(HttpResponse.ResponseHeadersEntry)
_sym_db.RegisterMessage(HttpResponse.NormalizedResponseHeadersEntry)

SimplePair = _reflection.GeneratedProtocolMessageType('SimplePair', (_message.Message,), {
  'DESCRIPTOR' : _SIMPLEPAIR,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.SimplePair)
  })
_sym_db.RegisterMessage(SimplePair)

Pair = _reflection.GeneratedProtocolMessageType('Pair', (_message.Message,), {
  'DESCRIPTOR' : _PAIR,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Pair)
  })
_sym_db.RegisterMessage(Pair)

AttackResult = _reflection.GeneratedProtocolMessageType('AttackResult', (_message.Message,), {
  'DESCRIPTOR' : _ATTACKRESULT,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.AttackResult)
  })
_sym_db.RegisterMessage(AttackResult)

RaspRuleSample = _reflection.GeneratedProtocolMessageType('RaspRuleSample', (_message.Message,), {
  'DESCRIPTOR' : _RASPRULESAMPLE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.RaspRuleSample)
  })
_sym_db.RegisterMessage(RaspRuleSample)

EndpointBanDetails = _reflection.GeneratedProtocolMessageType('EndpointBanDetails', (_message.Message,), {
  'DESCRIPTOR' : _ENDPOINTBANDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.EndpointBanDetails)
  })
_sym_db.RegisterMessage(EndpointBanDetails)

ModSecurityDetails = _reflection.GeneratedProtocolMessageType('ModSecurityDetails', (_message.Message,), {
  'DESCRIPTOR' : _MODSECURITYDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ModSecurityDetails)
  })
_sym_db.RegisterMessage(ModSecurityDetails)

ModSecurityLog = _reflection.GeneratedProtocolMessageType('ModSecurityLog', (_message.Message,), {

  'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
    'DESCRIPTOR' : _MODSECURITYLOG_PARAMSENTRY,
    '__module__' : 'dtm_pb2'
    # @@protoc_insertion_point(class_scope:contrast.api.dtm.ModSecurityLog.ParamsEntry)
    })
  ,
  'DESCRIPTOR' : _MODSECURITYLOG,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ModSecurityLog)
  })
_sym_db.RegisterMessage(ModSecurityLog)
_sym_db.RegisterMessage(ModSecurityLog.ParamsEntry)

BruteForceDetails = _reflection.GeneratedProtocolMessageType('BruteForceDetails', (_message.Message,), {
  'DESCRIPTOR' : _BRUTEFORCEDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.BruteForceDetails)
  })
_sym_db.RegisterMessage(BruteForceDetails)

BotBlockerDetails = _reflection.GeneratedProtocolMessageType('BotBlockerDetails', (_message.Message,), {
  'DESCRIPTOR' : _BOTBLOCKERDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.BotBlockerDetails)
  })
_sym_db.RegisterMessage(BotBlockerDetails)

CmdInjectionDetails = _reflection.GeneratedProtocolMessageType('CmdInjectionDetails', (_message.Message,), {
  'DESCRIPTOR' : _CMDINJECTIONDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.CmdInjectionDetails)
  })
_sym_db.RegisterMessage(CmdInjectionDetails)

CmdInjectionSemanticAnalysisDetails = _reflection.GeneratedProtocolMessageType('CmdInjectionSemanticAnalysisDetails', (_message.Message,), {
  'DESCRIPTOR' : _CMDINJECTIONSEMANTICANALYSISDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.CmdInjectionSemanticAnalysisDetails)
  })
_sym_db.RegisterMessage(CmdInjectionSemanticAnalysisDetails)

CsrfDetails = _reflection.GeneratedProtocolMessageType('CsrfDetails', (_message.Message,), {
  'DESCRIPTOR' : _CSRFDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.CsrfDetails)
  })
_sym_db.RegisterMessage(CsrfDetails)

CveDetails = _reflection.GeneratedProtocolMessageType('CveDetails', (_message.Message,), {
  'DESCRIPTOR' : _CVEDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.CveDetails)
  })
_sym_db.RegisterMessage(CveDetails)

Cve20170898Details = _reflection.GeneratedProtocolMessageType('Cve20170898Details', (_message.Message,), {
  'DESCRIPTOR' : _CVE20170898DETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Cve20170898Details)
  })
_sym_db.RegisterMessage(Cve20170898Details)

Cve201714033Details = _reflection.GeneratedProtocolMessageType('Cve201714033Details', (_message.Message,), {
  'DESCRIPTOR' : _CVE201714033DETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Cve201714033Details)
  })
_sym_db.RegisterMessage(Cve201714033Details)

Cve201714064Details = _reflection.GeneratedProtocolMessageType('Cve201714064Details', (_message.Message,), {
  'DESCRIPTOR' : _CVE201714064DETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Cve201714064Details)
  })
_sym_db.RegisterMessage(Cve201714064Details)

Cve201717405Details = _reflection.GeneratedProtocolMessageType('Cve201717405Details', (_message.Message,), {
  'DESCRIPTOR' : _CVE201717405DETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.Cve201717405Details)
  })
_sym_db.RegisterMessage(Cve201717405Details)

HttpMethodTamperingDetails = _reflection.GeneratedProtocolMessageType('HttpMethodTamperingDetails', (_message.Message,), {
  'DESCRIPTOR' : _HTTPMETHODTAMPERINGDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.HttpMethodTamperingDetails)
  })
_sym_db.RegisterMessage(HttpMethodTamperingDetails)

UntrustedDeserializationDetails = _reflection.GeneratedProtocolMessageType('UntrustedDeserializationDetails', (_message.Message,), {
  'DESCRIPTOR' : _UNTRUSTEDDESERIALIZATIONDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.UntrustedDeserializationDetails)
  })
_sym_db.RegisterMessage(UntrustedDeserializationDetails)

ElInjectionDetails = _reflection.GeneratedProtocolMessageType('ElInjectionDetails', (_message.Message,), {
  'DESCRIPTOR' : _ELINJECTIONDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ElInjectionDetails)
  })
_sym_db.RegisterMessage(ElInjectionDetails)

IpBlacklistDetails = _reflection.GeneratedProtocolMessageType('IpBlacklistDetails', (_message.Message,), {
  'DESCRIPTOR' : _IPBLACKLISTDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.IpBlacklistDetails)
  })
_sym_db.RegisterMessage(IpBlacklistDetails)

IpDenylistDetails = _reflection.GeneratedProtocolMessageType('IpDenylistDetails', (_message.Message,), {
  'DESCRIPTOR' : _IPDENYLISTDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.IpDenylistDetails)
  })
_sym_db.RegisterMessage(IpDenylistDetails)

MarkOfTheBeastDetails = _reflection.GeneratedProtocolMessageType('MarkOfTheBeastDetails', (_message.Message,), {
  'DESCRIPTOR' : _MARKOFTHEBEASTDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.MarkOfTheBeastDetails)
  })
_sym_db.RegisterMessage(MarkOfTheBeastDetails)

PaddingOracleDetails = _reflection.GeneratedProtocolMessageType('PaddingOracleDetails', (_message.Message,), {
  'DESCRIPTOR' : _PADDINGORACLEDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.PaddingOracleDetails)
  })
_sym_db.RegisterMessage(PaddingOracleDetails)

PathTraversalDetails = _reflection.GeneratedProtocolMessageType('PathTraversalDetails', (_message.Message,), {
  'DESCRIPTOR' : _PATHTRAVERSALDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.PathTraversalDetails)
  })
_sym_db.RegisterMessage(PathTraversalDetails)

PathTraversalSemanticAnalysisDetails = _reflection.GeneratedProtocolMessageType('PathTraversalSemanticAnalysisDetails', (_message.Message,), {
  'DESCRIPTOR' : _PATHTRAVERSALSEMANTICANALYSISDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.PathTraversalSemanticAnalysisDetails)
  })
_sym_db.RegisterMessage(PathTraversalSemanticAnalysisDetails)

ReDosDetails = _reflection.GeneratedProtocolMessageType('ReDosDetails', (_message.Message,), {
  'DESCRIPTOR' : _REDOSDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ReDosDetails)
  })
_sym_db.RegisterMessage(ReDosDetails)

SsjsInjectionDetails = _reflection.GeneratedProtocolMessageType('SsjsInjectionDetails', (_message.Message,), {
  'DESCRIPTOR' : _SSJSINJECTIONDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.SsjsInjectionDetails)
  })
_sym_db.RegisterMessage(SsjsInjectionDetails)

SqlInjectionDetails = _reflection.GeneratedProtocolMessageType('SqlInjectionDetails', (_message.Message,), {
  'DESCRIPTOR' : _SQLINJECTIONDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.SqlInjectionDetails)
  })
_sym_db.RegisterMessage(SqlInjectionDetails)

SqlInjectionSemanticAnalysisDetails = _reflection.GeneratedProtocolMessageType('SqlInjectionSemanticAnalysisDetails', (_message.Message,), {
  'DESCRIPTOR' : _SQLINJECTIONSEMANTICANALYSISDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.SqlInjectionSemanticAnalysisDetails)
  })
_sym_db.RegisterMessage(SqlInjectionSemanticAnalysisDetails)

NoSqlInjectionDetails = _reflection.GeneratedProtocolMessageType('NoSqlInjectionDetails', (_message.Message,), {
  'DESCRIPTOR' : _NOSQLINJECTIONDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.NoSqlInjectionDetails)
  })
_sym_db.RegisterMessage(NoSqlInjectionDetails)

SsrfDetails = _reflection.GeneratedProtocolMessageType('SsrfDetails', (_message.Message,), {
  'DESCRIPTOR' : _SSRFDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.SsrfDetails)
  })
_sym_db.RegisterMessage(SsrfDetails)

VirtualPatchDetails = _reflection.GeneratedProtocolMessageType('VirtualPatchDetails', (_message.Message,), {
  'DESCRIPTOR' : _VIRTUALPATCHDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.VirtualPatchDetails)
  })
_sym_db.RegisterMessage(VirtualPatchDetails)

XssDetails = _reflection.GeneratedProtocolMessageType('XssDetails', (_message.Message,), {
  'DESCRIPTOR' : _XSSDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.XssDetails)
  })
_sym_db.RegisterMessage(XssDetails)

XssMatch = _reflection.GeneratedProtocolMessageType('XssMatch', (_message.Message,), {
  'DESCRIPTOR' : _XSSMATCH,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.XssMatch)
  })
_sym_db.RegisterMessage(XssMatch)

XxeDetails = _reflection.GeneratedProtocolMessageType('XxeDetails', (_message.Message,), {
  'DESCRIPTOR' : _XXEDETAILS,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.XxeDetails)
  })
_sym_db.RegisterMessage(XxeDetails)

XxeMatch = _reflection.GeneratedProtocolMessageType('XxeMatch', (_message.Message,), {
  'DESCRIPTOR' : _XXEMATCH,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.XxeMatch)
  })
_sym_db.RegisterMessage(XxeMatch)

XxeWrapper = _reflection.GeneratedProtocolMessageType('XxeWrapper', (_message.Message,), {
  'DESCRIPTOR' : _XXEWRAPPER,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.XxeWrapper)
  })
_sym_db.RegisterMessage(XxeWrapper)

UserInput = _reflection.GeneratedProtocolMessageType('UserInput', (_message.Message,), {
  'DESCRIPTOR' : _USERINPUT,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.UserInput)
  })
_sym_db.RegisterMessage(UserInput)

StackTraceElement = _reflection.GeneratedProtocolMessageType('StackTraceElement', (_message.Message,), {
  'DESCRIPTOR' : _STACKTRACEELEMENT,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.StackTraceElement)
  })
_sym_db.RegisterMessage(StackTraceElement)

TraceEvent = _reflection.GeneratedProtocolMessageType('TraceEvent', (_message.Message,), {
  'DESCRIPTOR' : _TRACEEVENT,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.TraceEvent)
  })
_sym_db.RegisterMessage(TraceEvent)

TraceEventSignature = _reflection.GeneratedProtocolMessageType('TraceEventSignature', (_message.Message,), {
  'DESCRIPTOR' : _TRACEEVENTSIGNATURE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.TraceEventSignature)
  })
_sym_db.RegisterMessage(TraceEventSignature)

TraceEventSource = _reflection.GeneratedProtocolMessageType('TraceEventSource', (_message.Message,), {
  'DESCRIPTOR' : _TRACEEVENTSOURCE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.TraceEventSource)
  })
_sym_db.RegisterMessage(TraceEventSource)

TraceEventObject = _reflection.GeneratedProtocolMessageType('TraceEventObject', (_message.Message,), {
  'DESCRIPTOR' : _TRACEEVENTOBJECT,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.TraceEventObject)
  })
_sym_db.RegisterMessage(TraceEventObject)

TraceTaintRange = _reflection.GeneratedProtocolMessageType('TraceTaintRange', (_message.Message,), {
  'DESCRIPTOR' : _TRACETAINTRANGE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.TraceTaintRange)
  })
_sym_db.RegisterMessage(TraceTaintRange)

ParentObjectId = _reflection.GeneratedProtocolMessageType('ParentObjectId', (_message.Message,), {
  'DESCRIPTOR' : _PARENTOBJECTID,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ParentObjectId)
  })
_sym_db.RegisterMessage(ParentObjectId)

TraceStack = _reflection.GeneratedProtocolMessageType('TraceStack', (_message.Message,), {
  'DESCRIPTOR' : _TRACESTACK,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.TraceStack)
  })
_sym_db.RegisterMessage(TraceStack)

RouteCoverage = _reflection.GeneratedProtocolMessageType('RouteCoverage', (_message.Message,), {
  'DESCRIPTOR' : _ROUTECOVERAGE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.RouteCoverage)
  })
_sym_db.RegisterMessage(RouteCoverage)

ObservedRoute = _reflection.GeneratedProtocolMessageType('ObservedRoute', (_message.Message,), {
  'DESCRIPTOR' : _OBSERVEDROUTE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.ObservedRoute)
  })
_sym_db.RegisterMessage(ObservedRoute)

CmdiSinkRequest = _reflection.GeneratedProtocolMessageType('CmdiSinkRequest', (_message.Message,), {
  'DESCRIPTOR' : _CMDISINKREQUEST,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.CmdiSinkRequest)
  })
_sym_db.RegisterMessage(CmdiSinkRequest)

CmdiSinkResponse = _reflection.GeneratedProtocolMessageType('CmdiSinkResponse', (_message.Message,), {
  'DESCRIPTOR' : _CMDISINKRESPONSE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.CmdiSinkResponse)
  })
_sym_db.RegisterMessage(CmdiSinkResponse)

PtSinkRequest = _reflection.GeneratedProtocolMessageType('PtSinkRequest', (_message.Message,), {
  'DESCRIPTOR' : _PTSINKREQUEST,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.PtSinkRequest)
  })
_sym_db.RegisterMessage(PtSinkRequest)

PtSinkResponse = _reflection.GeneratedProtocolMessageType('PtSinkResponse', (_message.Message,), {
  'DESCRIPTOR' : _PTSINKRESPONSE,
  '__module__' : 'dtm_pb2'
  # @@protoc_insertion_point(class_scope:contrast.api.dtm.PtSinkResponse)
  })
_sym_db.RegisterMessage(PtSinkResponse)


DESCRIPTOR._options = None
_MESSAGE.fields_by_name['client_number']._options = None
_MESSAGE.fields_by_name['client_total']._options = None
_MESSAGE.fields_by_name['noop']._options = None
_NOOP._options = None
_SERVERACTIVITY_LOGENHANCERSENTRY._options = None
_SERVERACTIVITY_IPBLACKLISTSENTRY._options = None
_SERVERACTIVITY_IPDENYLISTSENTRY._options = None
_SERVERACTIVITY.fields_by_name['ip_blacklists']._options = None
_AGENTSTARTUP.fields_by_name['version']._options = None
_AGENTSTARTUP.fields_by_name['tags']._options = None
_APPLICATIONCREATE.fields_by_name['tags']._options = None
_APPLICATIONUPDATE_TECHNOLOGIESENTRY._options = None
_APPLICATIONUPDATE_LIBRARIESENTRY._options = None
_ACTIVITY_TECHNOLOGIESENTRY._options = None
_ACTIVITY_LIBRARYUSAGESENTRY._options = None
_ACTIVITY_DYNAMICSOURCESENTRY._options = None
_ACTIVITY.fields_by_name['finding_tags']._options = None
_FINDING_PROPERTIESENTRY._options = None
_FINDING.fields_by_name['tags']._options = None
_FINDING.fields_by_name['session_id']._options = None
_DYNAMICSOURCE_PROPERTIESENTRY._options = None
_LIBRARY.fields_by_name['tags']._options = None
_LIBRARYUSAGEUPDATE_CLASSNAMESENTRY._options = None
_HTTPREQUEST_NORMALIZEDREQUESTPARAMSENTRY._options = None
_HTTPREQUEST_REQUESTHEADERSENTRY._options = None
_HTTPREQUEST_NORMALIZEDREQUESTHEADERSENTRY._options = None
_HTTPREQUEST_NORMALIZEDCOOKIESENTRY._options = None
_HTTPREQUEST.fields_by_name['normalized_request_headers']._options = None
_HTTPREQUEST.fields_by_name['request_body']._options = None
_HTTPRESPONSE_RESPONSEHEADERSENTRY._options = None
_HTTPRESPONSE_NORMALIZEDRESPONSEHEADERSENTRY._options = None
_HTTPRESPONSE.fields_by_name['response_headers']._options = None
_HTTPRESPONSE.fields_by_name['parsed_response_headers']._options = None
_HTTPRESPONSE.fields_by_name['response_body']._options = None
_ATTACKRESULT.fields_by_name['tags']._options = None
_RASPRULESAMPLE.fields_by_name['cve']._options = None
_RASPRULESAMPLE.fields_by_name['ip_blacklist']._options = None
_RASPRULESAMPLE.fields_by_name['cve_2017_0898']._options = None
_RASPRULESAMPLE.fields_by_name['cve_2017_14033']._options = None
_RASPRULESAMPLE.fields_by_name['cve_2017_14064']._options = None
_RASPRULESAMPLE.fields_by_name['cve_2017_17405']._options = None
_MODSECURITYLOG_PARAMSENTRY._options = None
_IPBLACKLISTDETAILS._options = None
_OBSERVEDROUTE.fields_by_name['session_id']._options = None
# @@protoc_insertion_point(module_scope)
